from __future__ import annotations
import sys
import itertools
from abc import ABC, abstractmethod
from collections import Counter
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

import numpy as np

from stereomolgraph.coords import are_planar, handedness, angle_from_coords

if TYPE_CHECKING:
    from collections.abc import Generator, Set

    from typing_extensions import Self

    from stereomolgraph.graphs.mg import AtomId, Bond

if sys.version_info >= (3, 13):
    from typing import TypeVar
else:
    from typing_extensions import TypeVar


A = TypeVar("A", bound=tuple[int, ...],
            covariant=True,
            default=tuple[int, ...])
P = TypeVar("P", bound=None | Literal[1, 0, -1],
            covariant=True,
            default=None | Literal[1, 0, -1])


class Stereo(ABC, Generic[A, P]):
    """
    :class:`~typing.Generic` Class to represent the orientation of a group of
    atoms in space.
    This is used to represent local stereochemistry and simultanously the
    hybridization of atoms.
    """

    atoms: A
    """Atoms are a order dependent tuple of integers."""

    parity: P
    """parity is a number that defines the orientation of the atoms. If None,
    the relative orientation of the atoms is not defined.
    If 0 the orientation is defined and part of a achiral stereochemistry.
    If 1 or -1 the orientation is defined and part of a chiral stereochemistry.
    """

    PERMUTATION_GROUP: frozenset[A]
    """Defines all allowed permutations defined by the symmetry group under
    which the stereochemistry is invariant."""

    @abstractmethod
    def __init__(self, atoms: A, parity: P = None): ...

    @abstractmethod
    def __eq__(self, other: Any) -> bool: ...

    @abstractmethod
    def __hash__(self) -> int: ...

    @abstractmethod
    def invert(self) -> Self:
        """Inverts the stereo. If the stereo is achiral, it returns itself."""

    @abstractmethod
    def get_isomers(self) -> Set[Self]:
        """Returns all stereoisomers of the stereochemistry. Not just the
        inverted ones, but all possible stereoisomers."""


class AtomStereo(Stereo[A, P], ABC, Generic[A, P]):
    @property
    def central_atom(self) -> AtomId:
        return self.atoms[0]


class BondStereo(Stereo[A, P], ABC, Generic[A, P]):
    @property
    def bond(self) -> Bond:
        bond = frozenset(self.atoms[2:4])
        assert len(bond) == 2
        return bond


class _StereoMixin(Stereo[A, P], ABC, Generic[A, P]):
    __slots__ = ("atoms", "parity")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.atoms}, {self.parity})"

    def __init__(self, atoms: A, parity: P = None):
        self.atoms = atoms
        self.parity = parity

    def _perm_atoms(self) -> Generator[tuple[int, ...], None, None]:
        if self.parity is None:
            return (
                tuple([self.atoms[i] for i in perm])
                for perm in itertools.permutations(range(len(self.atoms)))
            )
        else:
            return (
                tuple([self.atoms[i] for i in perm])
                for perm in self.PERMUTATION_GROUP
            )


class _ChiralStereoMixin(
    _StereoMixin[A, None | Literal[1, -1]], ABC, Generic[A]
):
    __slots__ = ("inversion",)
    inversion: A

    def invert(self) -> Self:
        if self.parity is None:
            return self
        return self.__class__(self.atoms, self.parity * -1)

    def _inverted_atoms(self) -> A:
        atoms = tuple([self.atoms[i] for i in self.inversion])
        assert len(atoms) == len(self.atoms) == len(self.inversion)
        return atoms  # type: ignore[return-value]

    def __eq__(self, other: Any) -> bool:
        if other.parity == 0:
            return False

        s_atoms, o_atoms = self.atoms, other.atoms
        set_s_atoms = set(s_atoms)
        set_o_atoms = set(o_atoms)

        if len(s_atoms) != len(o_atoms) or not set_s_atoms.issuperset(
            set_o_atoms
        ):
            return False

        if self.parity is None or other.parity is None:
            if set_s_atoms == set_o_atoms:
                return True
            return False

        elif self.parity == other.parity:
            if o_atoms == s_atoms or any(
                o_atoms == p for p in self._perm_atoms()
            ):
                return True
            return False

        elif self.parity * -1 == other.parity:
            if any(other._inverted_atoms() == p for p in self._perm_atoms()):
                return True
            return False

        raise RuntimeError("This should not happen")

    def __hash__(self) -> int:
        if self.parity is None:
            return hash(frozenset(Counter(self.atoms).items()))
        perm = frozenset(
            {
                tuple([self.atoms[i] for i in perm])
                for perm in self.PERMUTATION_GROUP
            }
        )

        inverted_perm = frozenset(
            {
                tuple([self._inverted_atoms()[i] for i in perm])
                for perm in self.PERMUTATION_GROUP
            }
        )

        if self.parity == 1:
            return hash((perm, inverted_perm))
        elif self.parity == -1:
            return hash((inverted_perm, perm))
        else:
            raise RuntimeError("Something is wrong with parity")


class _AchiralStereoMixin(_StereoMixin[A, None | Literal[0]], ABC, Generic[A]):
    __slots__: tuple[str, ...] = ()

    def invert(self) -> Self:
        return self

    def __eq__(self, other: Any) -> bool:
        if other.parity in (1, -1):
            return False

        s_atoms, o_atoms = self.atoms, other.atoms
        set_s_atoms = set(s_atoms)
        set_o_atoms = set(o_atoms)

        if len(s_atoms) != len(o_atoms) or not set_s_atoms.issuperset(
            set_o_atoms
        ):
            return False

        if self.parity is None or other.parity is None:
            if set_s_atoms == set_o_atoms:
                return True
            return False

        if self.parity == other.parity:
            if o_atoms == s_atoms or o_atoms in self._perm_atoms():
                return True
            return False

        raise RuntimeError("This should not happen!")

    def __hash__(self) -> int:
        if self.parity is None:
            return hash(frozenset(Counter(self.atoms).items()))
        elif self.parity == 0:
            perm = frozenset(
                {
                    tuple([self.atoms[i] for i in perm])
                    for perm in self.PERMUTATION_GROUP
                }
            )
            return hash(perm)
        raise RuntimeError("Something is wrong with parity")


class Tetrahedral(
    _ChiralStereoMixin[tuple[int, int, int, int, int]],
    AtomStereo[tuple[int, int, int, int, int], None | Literal[1, -1]],
):
    r"""Represents all possible configurations of atoms for a Tetrahedral
    Stereochemistry::

       parity = 1      parity = -1
           4                4
           |                |
           0                0
        /  ¦  \          /  ¦  \
       2   1   3        3   1   2

    Atoms of the tetrahedral stereochemistry are ordered in a way that when the
    first atom is rotated to the back, the other atoms in order are rotated in
    the direction defined by the stereo.

    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    :ivar PERMUTATION_GROUP: Permutations allowed by the stereochemistry
    """

    __slots__ = ()
    inversion = (0, 2, 1, 3, 4)
    PERMUTATION_GROUP = frozenset(
        {
            (0, 1, 2, 3, 4),
            (0, 3, 1, 2, 4),
            (0, 2, 3, 1, 4),
            (0, 1, 4, 2, 3),
            (0, 2, 1, 4, 3),
            (0, 4, 2, 1, 3),
            (0, 1, 3, 4, 2),
            (0, 4, 1, 3, 2),
            (0, 3, 4, 1, 2),
            (0, 2, 4, 3, 1),
            (0, 3, 2, 4, 1),
            (0, 4, 3, 2, 1),
        }
    )

    def __init__(
        self,
        atoms: tuple[int, int, int, int, int],
        parity: None | Literal[1, -1] = None,
    ):
        super().__init__(atoms=atoms, parity=parity)

    def get_isomers(self) -> set[Self]:
        return {
            self.__class__(atoms=self.atoms, parity=1),
            self.__class__(atoms=self.atoms, parity=-1),
        }

    @classmethod
    def from_coords(
        cls,
        atoms: tuple[int, int, int, int, int],
        coords: np.ndarray[
            tuple[Literal[5], Literal[3]], np.dtype[np.float64]
        ],
    ) -> Self:
        """
        Creates the representation of a Tetrahedral Stereochemistry
        from the coordinates of the atoms.

        :param atoms: Atoms of the stereochemistry
        :param coords: nAtomsx3 numpy array with cartesian coordinates
        """
        orientation = handedness(coords.take((1, 2, 3, 4), axis=0))
        int_orientation = int(orientation)
        assert int_orientation in (1, -1), (
            f"Orientation {orientation} is not valid for Tetrahedral "
            "stereochemistry.")
        return cls(atoms, int_orientation)


class SquarePlanar(
    _AchiralStereoMixin[tuple[int, int, int, int, int]],
    AtomStereo[tuple[int, int, int, int, int], None | Literal[0]],
):
    r""" Represents all possible configurations of atoms for a
    SquarePlanar Stereochemistry::

        1     4
         \   /
           0
         /   \
        2     3

    Atoms of the Square Planar stereochemistry are ordered in a way that


    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    """

    __slots__ = ()
    PERMUTATION_GROUP = frozenset(
        {
            (0, 1, 2, 3, 4),
            (0, 2, 3, 4, 1),
            (0, 3, 4, 1, 2),
            (0, 4, 1, 2, 3),
            (0, 4, 3, 2, 1),
            (0, 3, 2, 1, 4),
            (0, 2, 1, 4, 3),
            (0, 1, 4, 3, 2),
        }
    )

    @property
    def central_atom(self) -> AtomId:
        return self.atoms[0]

    def get_isomers(self) -> set[SquarePlanar]:
        return {
            SquarePlanar(atoms=atoms, parity=0)
            for perm in itertools.permutations(self.atoms[1:])
            if len(atoms := (self.atoms[0], *perm)) == 5
        }


class TrigonalBipyramidal(
    _ChiralStereoMixin[tuple[int, int, int, int, int, int]],
    AtomStereo[tuple[int, int, int, int, int, int], None | Literal[1, -1]],
):
    r"""Represents all possible configurations of atoms for a
    TrigonalBipyramidal Stereochemistry::

       parity = 1             parity = -1
        3   1                     1   3
         ◁  ¦                    ¦  ▷
            0  — 5           5 —  0
         ◀  ¦                    ¦  ▶
        4   2                     2   4

    Atoms of the trigonal bipyramidal stereochemistry are ordered in a way that
    when the first two atoms are the top and bottom of the bipyramid. The last
    three equatorial atoms are ordered in a way that when the first atom is
    rotated to the back, the other atoms in order are rotated in the direction
    defined by the stereo.

    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    """

    __slots__ = ()
    inversion = (0, 1, 2, 3, 5, 4)
    PERMUTATION_GROUP = frozenset(
        {
            (0, 1, 2, 3, 4, 5),
            (0, 1, 2, 5, 3, 4),
            (0, 1, 2, 4, 5, 3),
            (0, 2, 1, 3, 5, 4),
            (0, 2, 1, 5, 4, 3),
            (0, 2, 1, 4, 3, 5),
        }
    )

    def get_isomers(self) -> set[Self]:
        return {
            self.__class__(atoms=atoms, parity=p)
            for perm in itertools.permutations(self.atoms[1:])
            for p in (1, -1)
            if len(atoms := (self.atoms[0], *perm)) == 6
        }

    @classmethod
    def from_coords(
        cls: type[TrigonalBipyramidal],
        atoms: tuple[int, int, int, int, int, int],
        coords: np.ndarray[
            tuple[Literal[6], Literal[3]], np.dtype[np.float64]
        ],
    ) -> TrigonalBipyramidal:
        """
        calculates the distance of the atom 5 from the plane defined by the
        first three atoms in Angstrom. The sign of the distance is determined
        by the side of the plane that atom 5 is on.
        """

        # coords_dict
        _index_central_atom = 0
        indices = (1, 2, 3, 4, 5)

        if np.any(are_planar(coords[[1,2,3,4,5]])):
            raise ValueError("Four atoms are planar!")
        
        lst = np.array([[i, 0, j] for i, j
                        in itertools.combinations(indices, 2)], dtype=np.int8)

        # The atoms with the largest angle are the axial atoms
        angles = angle_from_coords(coords[lst])
        i, j = lst[angles.argmax()][[0, 2]] # axial atoms
        
        equatorial = [a for a in indices if a not in (i, j)]
        i_rotation = handedness(coords.take( [*equatorial, i], axis=0))
        j_rotation = -1 * handedness(coords.take([*equatorial, j], axis=0))

        assert int(i_rotation) == int(j_rotation)

        atoms_in_new_order = (i, j, *equatorial)
        orientation = int(i_rotation)
        tb_atoms = (atoms[0], *atoms_in_new_order)
        assert len(tb_atoms) == 6
        assert orientation in (1, -1)
        return TrigonalBipyramidal(tb_atoms, orientation)
        
class Octahedral(
    _ChiralStereoMixin[tuple[int, int, int, int, int, int, int]],
    AtomStereo[
        tuple[int, int, int, int, int, int, int], None | Literal[1, -1]
    ],
):
    """Represents all possible configurations of atoms for a Octahedral
    Stereochemistry::

        parity = 1             parity = -1
         3  1   6                3  2  6
          ◁ ¦ /                  ◁ ¦ /
            0                       0
          / ¦ ▶                  / ¦  ▶
         4  2  5                4   1  5
    """

    __slots__ = ()
    inversion = (0, 2, 1, 3, 4, 5, 6)
    PERMUTATION_GROUP = frozenset(
        {
            (0, 1, 2, 3, 4, 5, 6),
            (0, 1, 2, 6, 3, 4, 5),
            (0, 1, 2, 5, 6, 3, 4),
            (0, 1, 2, 4, 5, 6, 3),
            (0, 2, 1, 4, 3, 6, 5),
            (0, 2, 1, 5, 4, 3, 6),
            (0, 2, 1, 6, 5, 4, 3),
            (0, 2, 1, 3, 6, 5, 4),
            (0, 3, 5, 2, 4, 1, 6),
            (0, 3, 5, 6, 2, 4, 1),
            (0, 3, 5, 1, 6, 2, 4),
            (0, 3, 5, 4, 1, 6, 2),
            (0, 5, 3, 1, 4, 2, 6),
            (0, 5, 3, 6, 1, 4, 2),
            (0, 5, 3, 2, 6, 1, 4),
            (0, 5, 3, 4, 2, 6, 1),
            (0, 4, 6, 3, 2, 5, 1),
            (0, 4, 6, 1, 3, 2, 5),
            (0, 4, 6, 5, 1, 3, 2),
            (0, 4, 6, 2, 5, 1, 3),
            (0, 6, 4, 3, 1, 5, 2),
            (0, 6, 4, 2, 3, 1, 5),
            (0, 6, 4, 5, 2, 3, 1),
            (0, 6, 4, 1, 5, 2, 3),
        }
    )

    def get_isomers(self) -> set[Octahedral]:
        return {
            Octahedral(atoms=atoms, parity=p)
            for perm in itertools.permutations(self.atoms[1:])
            for p in (1, -1)
            if len((atoms := (self.atoms[0], *perm))) == 7
        }


class PlanarBond(
    _AchiralStereoMixin[tuple[int, int, int, int, int, int]],
    BondStereo[tuple[int, int, int, int, int, int], None | Literal[0]],
):
    r""" Represents all possible configurations of atoms for a
    Planar Structure and should be used for aromatic and double bonds::

        0        4
         \      /
          2 == 3
         /      \
        1        5

    All atoms of the double bond are in one plane. Atoms 2 and 3 are the center
    Atoms 0 and 1 are bonded to 2 and atoms 4 and 5 are bonded to 3.
    The stereochemistry is defined by the relative orientation
    of the atoms 0, 1, 4 and 5.

    :ivar atoms: Atoms of the stereochemistry
    :ivar parity: Stereochemistry
    :ivar PERMUTATION_GROUP: Permutations allowed by the stereochemistry
    """

    __slots__ = ()
    PERMUTATION_GROUP = frozenset(
        {
            (0, 1, 2, 3, 4, 5),
            (1, 0, 2, 3, 5, 4),
            (4, 5, 3, 2, 0, 1),
            (5, 4, 3, 2, 1, 0),
        }
    )

    def get_isomers(self) -> set[PlanarBond]:
        reordered_atoms = tuple(self.atoms[i] for i in (0, 1, 2, 3, 5, 4))
        assert len(reordered_atoms) == 6
        return {
            PlanarBond(self.atoms, 0),
            PlanarBond(reordered_atoms, 0),
        }

    def __init__(
        self,
        atoms: tuple[int, int, int, int, int, int],
        parity: None | Literal[0] = None,
    ):
        super().__init__(atoms, parity)

    @classmethod
    def from_coords(
        cls,
        atoms: tuple[int, int, int, int, int, int],
        coords: np.ndarray[
            tuple[Literal[6], Literal[3]], np.dtype[np.float64]
        ],
    ) -> PlanarBond:
        a = (coords[0] - coords[1]) / np.linalg.norm(coords[0] - coords[1])
        b = (coords[4] - coords[5]) / np.linalg.norm(coords[4] - coords[5])
        result = int(np.sign(np.dot(a, b)))

        if result == -1:
            new_atoms = tuple(atoms[i] for i in (1, 0, 2, 3, 4, 5))
        elif result == 1:
            new_atoms = atoms
        elif result == 0:
            raise ValueError("atoms are tetrahedral")
        else:
            raise ValueError("something went wrong")
        assert len(new_atoms) == 6
        return cls(new_atoms, 0)


class AtropBond(
    _ChiralStereoMixin[tuple[int, int, int, int, int, int]],
    BondStereo[tuple[int, int, int, int, int, int], None | Literal[1, -1]],
):
    r"""
    Represents all possible configurations of atoms for a
    Atropostereoisomer bond::
    
        parity = 1          parity = -1
        1       5           1        5
         \     /            ◀      /
          2 - 3               2 - 3
        ◀      \            /      \
        0        4         0         4


    """

    __slots__ = ()
    inversion = (1, 0, 2, 3, 4, 5)
    PERMUTATION_GROUP = frozenset(
        {
            (0, 1, 2, 3, 4, 5),
            (1, 0, 2, 3, 5, 4),
            (4, 5, 3, 2, 1, 0),
            (5, 4, 3, 2, 0, 1),
        }
    )

    def get_isomers(self) -> set[AtropBond]:
        other_atoms = tuple(self.atoms[i] for i in (0, 1, 2, 3, 5, 4))
        assert len(other_atoms) == 6
        return {
            AtropBond(self.atoms, 1),
            AtropBond(other_atoms, -1),
        }



# pyright: standard
from __future__ import annotations

from rdkit import Chem # type: ignore
from rdkit.Chem import Draw # type: ignore
from typing import NamedTuple

from stereomolgraph import (
        MolGraph,
        CondensedReactionGraph,
        StereoMolGraph,
        StereoCondensedReactionGraph)
from stereomolgraph.stereodescriptors import PlanarBond

def default_repr_svg(graph):
    return View2D().svg(graph)

def default_view_molgraph(self: MolGraph) -> None:
    View2D()(self)

MolGraph._ipython_display_ = default_view_molgraph


class View2D(NamedTuple):
    height: int = 300
    width: int = 300
    show_atom_numbers: bool = True
    show_h: bool = True
    generate_bond_orders: bool = False
    dummy_atoms: bool = True

    def svg(
        self,
        graph: (
            MolGraph
            | CondensedReactionGraph
            | StereoMolGraph
            | StereoCondensedReactionGraph
        ),
    ) -> str:
        mol, idx_map_num_dict = graph._to_rdmol(
            generate_bond_orders=self.generate_bond_orders
        )
        map_num_idx_dict = {v: k for k, v in idx_map_num_dict.items()}
        
        if not self.generate_bond_orders:
            for bond in mol.GetBonds():
                bond.SetBondType(Chem.BondType.SINGLE)

        if self.show_atom_numbers:
            for atom in mol.GetAtoms():
                atom.SetProp("atomNote", str(atom.GetAtomMapNum()))
                atom.SetAtomMapNum(0)
        else:
            for atom in mol.GetAtoms():
                atom.SetAtomMapNum(0)

        atoms_to_highlight = []
        highlight_atom_colors = {}

        bonds_to_highlight = []
        highlight_bond_colors = {}

        if self.dummy_atoms is False:
            dummy_atoms = [
                atom.GetIdx()
                for atom in mol.GetAtoms()
                if atom.GetSymbol() == "*"
            ]
            dummy_atoms.sort(reverse=True)
            for atom in dummy_atoms:
                mol.RemoveAtom(atom)

        if not self.show_h:
            mol = Chem.RemoveHs(mol, implicitOnly=False, sanitize=False)

        if isinstance(graph, StereoMolGraph):
            for db in graph.bond_stereo.values():
                if not isinstance(db, PlanarBond):
                    a1 = map_num_idx_dict[db.atoms[2]]
                    a2 = map_num_idx_dict[db.atoms[3]]
                    rd_bond = mol.GetBondBetweenAtoms(a1, a2)
                    rd_bond.SetBondType(Chem.BondType.AROMATIC)

        if isinstance(graph, CondensedReactionGraph):
            for bond in graph.get_formed_bonds():
                atoms_idx = [map_num_idx_dict[a] for a in bond]
                bond_idx = mol.GetBondBetweenAtoms(*atoms_idx).GetIdx()
                bonds_to_highlight.append(bond_idx)
                mol.GetBondWithIdx(bond_idx).SetBondType(
                    Chem.rdchem.BondType.HYDROGEN)
                highlight_bond_colors[bond_idx] = (0, 0, 1)  # blue

            for bond in graph.get_broken_bonds():
                atoms_idx = [map_num_idx_dict[a] for a in bond]
                bond_idx = mol.GetBondBetweenAtoms(*atoms_idx).GetIdx()
                bonds_to_highlight.append(bond_idx)
                mol.GetBondWithIdx(bond_idx).SetBondType(
                    Chem.rdchem.BondType.HYDROGEN)
                highlight_bond_colors[bond_idx] = (1, 0, 0)  # red

            # make dummy atoms and their bonds grey
        if self.dummy_atoms is True:
            atom_colors = {}
            grey = (0.7, 0.7, 0.7)
            for atom in mol.GetAtoms():
                if atom.GetSymbol() == "*":
                    atom.SetProp("atomLabel", "")
                    atom_colors[atom.GetIdx()] = grey
                    for bond in atom.GetBonds():
                        bonds_to_highlight.append(bond.GetIdx())
                        highlight_bond_colors[bond.GetIdx()] = grey

        Chem.rdDepictor.Compute2DCoords(mol, useRingTemplates=True) # type: ignore
        Chem.rdDepictor.StraightenDepiction(mol) # type: ignore

        drawer = Draw.rdMolDraw2D.MolDraw2DSVG(self.width, self.height)

        drawer.drawOptions().useBWAtomPalette()
        drawer.drawOptions().continuousHighlight = False
        drawer.drawOptions().highlightBondWidthMultiplier = 10
        drawer.drawOptions().fillHighlights = False
        drawer.drawOptions().includeRadicals = False

        drawer.DrawMolecule(
            mol,
            highlightAtoms=atoms_to_highlight,
            highlightAtomColors=highlight_atom_colors,
            highlightBonds=bonds_to_highlight,
            highlightBondColors=highlight_bond_colors,
        )

        drawer.FinishDrawing()
        svg = drawer.GetDrawingText()
        return svg
        
    def __call__(
        self,
        graph: (
            MolGraph
            | CondensedReactionGraph
            | StereoMolGraph
            | StereoCondensedReactionGraph
        ),
    ):
        # imported here, so that this module does not depend on IPython
        from IPython.display import SVG
        svg = self.svg(graph)
        display(SVG(svg.replace("svg:", "")))  # type: ignore # noqa



from __future__ import annotations

from collections import deque
from itertools import combinations
from typing import TYPE_CHECKING

import numpy as np

from stereomolgraph import (
    COVALENT_RADII,
    PERIODIC_TABLE,
    Element,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping, Sequence
    from os import PathLike
    from typing import Literal, TypeVar

    NP_FLOAT = TypeVar(
        "NP_FLOAT", bound=np.dtype[np.floating], contravariant=True
    )
    N = TypeVar("N", bound=int)
    ONE = Literal[1]
    #TWO = Literal[2]
    THREE = Literal[3]
    FOUR = Literal[4]

def are_planar(points: np.ndarray[tuple[int, Literal[3]], np.dtype[np.float64]],
               threshold: float = 0.5
               ) -> np.bool_:
    """Checks if all atoms are in one plane

    Checks if the all atoms are planar within a given threshold.
    The threshold is the maximal distance of an atom from the plane of
    three other atoms.

    :param points: coordinates of the atoms
    :type points: np.ndarray
    :param threshold: maximal distance of atom from plane [Angstrom]
    :type threshold: float
    :return: True if all atoms are planar
    :rtype: bool
    """
    if threshold <= 0:
        raise ValueError("threshold has to be bigger than 0")
    if len(points) < 4:
        return np.bool_(True)

    for p1, p2, p3, p4 in combinations(points, 4):
        d = deque([p1, p2, p3, p4])
        for _ in range(4):
            d.rotate()
            vec1 = p1 - p2
            vec2 = p3 - p2
            vec3 = p4 - p2
            normal = np.cross(vec1, vec2)
            norm_normal = normal / np.linalg.norm(normal)
            result = abs(np.dot(norm_normal, vec3))
            if result > threshold:
                return np.bool_(False)
    return np.bool_(True)

def are_planar_volume(
    coords: np.ndarray[tuple[FOUR, THREE], NP_FLOAT], threshold: float = 3.0
) -> np.ndarray[tuple[ONE], np.dtype[np.bool_]]:
    if coords.shape[-2] > 4:
        raise NotImplementedError(
            "are_planar_volume is not implemented for more than 4 points"
        )
        #for comb in combinations(range(coords.shape[-2]), 4):
        #    if not are_planar_volume(coords[..., list(comb), :], threshold=threshold):
        #        return np.array([False], dtype=np.bool_)
        #    else:
        #        return np.array([True], dtype=np.bool_)

    assert coords.shape[-1] == 3
    assert coords.shape[-2] == 4

    v1 = coords[..., 0, :] - coords[..., 1, :]
    v2 = coords[..., 2, :] - coords[..., 3, :]
    v3 = coords[..., 3, :] - coords[..., 1, :]

    normal = np.cross(v2, v3, axis=-1)

    volumes = np.abs(np.sum(v1 * normal, axis=-1))

    return volumes < threshold


def handedness(
    coords: np.ndarray[tuple[FOUR, THREE], NP_FLOAT],
) -> np.ndarray[tuple[ONE], np.dtype[np.int8]]:
    """
    Calculates the orientation of the atom 4 from the plane defined
    by the first three atoms from their coordinates.


    """
    vec1 = coords[..., 0, :] - coords[..., 1, :]
    vec2 = coords[..., 2, :] - coords[..., 3, :]
    vec3 = coords[..., 3, :] - coords[..., 1, :]

    normal = np.cross(vec1, vec2, axis=-1)

    # Normalize normal vectors (in-place operations)
    norm = np.linalg.norm(normal, axis=-1, keepdims=True)
    np.divide(normal, norm, out=normal)  # in-place division

    # Compute dot product (in-place multiply and sum)
    np.multiply(normal, vec3, out=normal)
    dot_product = np.sum(normal, axis=-1)
    result = np.sign(dot_product).astype(np.int8)
    return result


def angle_from_coords(
    coords: np.ndarray[tuple[THREE, THREE], NP_FLOAT],
    out: None | np.ndarray[tuple[ONE], NP_FLOAT] = None,
) -> np.ndarray[tuple[int, ...], NP_FLOAT]:
    """ """
    assert np.issubdtype(coords.dtype, np.floating)
    assert coords.shape[-1] == 3
    assert coords.shape[-2] == 3

    BA = coords[..., 0, :] - coords[..., 1, :]
    BC = coords[..., 2, :] - coords[..., 1, :]

    # Compute dot product and norms
    dot_product = np.einsum("...i,...i->...", BA, BC)
    assert dot_product is not None
    norm_BA = np.linalg.norm(BA, axis=-1)
    norm_BC = np.linalg.norm(BC, axis=-1)

    if out is not None:
        assert out.shape == dot_product.shape, "Output array must match shape"
        assert out.dtype == coords.dtype, "Output array must match dtype"
    else:
        output_shape = coords.shape[:-2]
        if output_shape == ():
            output_shape = (1,)

        out = np.empty(output_shape, dtype=coords.dtype)  # type: ignore
        assert out is not None, "Output array must be provided or created"
        assert out.shape == output_shape, (
            "Output array must be created with correct shape"
        )

    # Compute cosine of angle with safe division
    denominator = np.multiply(norm_BA, norm_BC, out=out)  # Reuse out as temp
    np.divide(dot_product, denominator, where=(denominator != 0), out=out)

    # Clip and compute angle
    np.clip(out, -1.0, 1.0, out=out)
    np.arccos(out, out=out)
    np.rad2deg(out, out=out)
    assert out is not None
    return out  # .squeeze()[()] if out.ndim == 0 else out


def pairwise_distances(
    coords: np.ndarray[tuple[N, THREE], NP_FLOAT],
) -> np.ndarray[tuple[N, N], NP_FLOAT]:
    if coords.shape[-1] != 3:
        raise ValueError("Last dimension must be size 3 for 3D coordinates")

    # Compute differences using broadcasting
    diff = (
        coords[..., :, None, :] - coords[..., None, :, :]
    )  # shape (..., N, N, 3)

    # Square differences and sum along last dimension
    np.square(diff, out=diff)  # in-place squaring
    summed = np.sum(diff, axis=-1)  # shape (..., N, N)

    # Compute distances in-place
    distances = np.sqrt(summed, out=summed)
    return distances


class Geometry:
    """
    Represents a molecular geometry, i.e. the coordinates and atom types.

    :param atom_types: tuple of Element objects
    :param coords: nAtomsx3 numpy array with cartesian coordinates
    """

    atom_types: tuple[Element, ...]
    coords: np.ndarray[tuple[int, Literal[3]], np.dtype[np.float64]]

    @property
    def n_atoms(self) -> int:
        return len(self.atom_types)

    def __len__(self) -> int:
        return self.n_atoms

    def __init__(
        self,
        atom_types: Iterable[int | str | Element] = tuple(),
        coords: np.ndarray[tuple[int, int], NP_FLOAT] = np.empty(
            (0, 3), dtype=np.float64
        ),
    ):
        self.coords = np.array(coords, dtype=np.float64)
        self.atom_types = tuple([PERIODIC_TABLE[atom] for atom in atom_types])

        assert (
            len(self.coords.shape) == 2
            and self.coords.shape[1] == 3
            and len(self.atom_types) == self.coords.shape[0]
        )

    @classmethod
    def from_xyz_file(cls, path: PathLike[str] | str) -> Geometry:
        """Create a Geometry from an XYZ file."""

        dt = np.dtype(
            [
                ("atom", "U5"),  # Unicode string up to 5 characters
                ("x", "f8"),  # 64-bit float
                ("y", "f8"),
                ("z", "f8"),
            ]
        )

        data = np.loadtxt(path, skiprows=2, dtype=dt, comments=None)

        atom_types = [PERIODIC_TABLE[atom] for atom in data["atom"]]
        coords = np.column_stack((data["x"], data["y"], data["z"]))

        return cls(atom_types=atom_types, coords=coords)

    def xyz_str(self, comment: None | str = None) -> str:
        """
        returns the xyz representation of this geometry as a string

        :param comment: comment for 2nd line of xyz file
        :return: xyz representation of this geometry
        """
        xyz = str(self.n_atoms) + "\n"
        if comment is not None:
            xyz += comment + "\n"
        else:
            xyz += "\n"

        for type, coords in zip(self.atom_types, self.coords):
            xyz += (
                f"{type.symbol:s} {coords[0]:.8f} {coords[1]:.8f} "
                f"{coords[2]:.8f}\n"
            )

        return xyz


def default_connectivity_cutoff(atom_types: tuple[Element, Element]) -> float:
    return sum(COVALENT_RADII[a] for a in atom_types) * 1.2


CONNECTIVITY_CUTOFF_FUNC: Callable[[tuple[Element, Element]], float] = (
    default_connectivity_cutoff
)


class _DefaultFuncDict(dict[tuple[Element, Element], float]):
    """
    A dictionary that calls a default function with keys as arguments,
    when a key is missing.
    """

    default_func: Callable[[tuple[Element, Element]], float]

    def __init__(
        self,
        *,
        default_func: Callable[
            [tuple[Element, Element]], float
        ] = CONNECTIVITY_CUTOFF_FUNC,
        **kwargs: Mapping[tuple[Element, Element], float],
    ):
        super().__init__(**kwargs)
        self.default_func = default_func

    def __missing__(self, key: tuple[Element, Element]) -> float:
        if len(key) != 2:
            raise KeyError(f"Key {key} must be a tuple of two elements.")
        elif (ret := self.get((key[1], key[0]), None)) is not None:
            pass
        else:
            ret = self.default_func(key)

        self[key] = ret

        return ret

    def array(self, atom_types: Sequence[Element]) -> np.ndarray:
        n_atoms = len(atom_types)
        array = np.zeros((n_atoms, n_atoms))
        for (atom1, atom_type1), (atom2, atom_type2) in combinations(
            enumerate(atom_types), 2
        ):
            value = self[atom_type1, atom_type2]
            array[atom1][atom2] = value
            array[atom2][atom1] = value
        return array


class BondsFromDistance:
    def __init__(
        self,
        connectivity_cutoff: Callable[
            [tuple[Element, Element]], float
        ] = CONNECTIVITY_CUTOFF_FUNC,
    ):
        self.connectivity_cutoff = _DefaultFuncDict(
            default_func=connectivity_cutoff
        )

    def __call__(
        self, distance: float, atom_types: tuple[Element, Element]
    ) -> Literal[0, 1]:
        if distance < 0:
            raise ValueError("distance can not be negative")
        else:
            return 1 if distance < self.connectivity_cutoff[atom_types] else 0

    def array(
        self,
        coords: np.ndarray[tuple[int, Literal[3]], np.dtype[np.float64]],
        atom_types: Sequence[Element],
    ) -> np.ndarray:
        return np.where(
            pairwise_distances(coords)
            < self.connectivity_cutoff.array(atom_types),
            1,
            0,
        )



"""Implementation of the isomorphism and subgraph isomorphism algorithms.
Based of VF2++ from https://doi.org/10.1016/j.dam.2018.02.018"""

from __future__ import annotations

from collections import Counter, defaultdict
from typing import TYPE_CHECKING, NamedTuple

from stereomolgraph.algorithms.color_refine import color_refine_mg, label_hash

if TYPE_CHECKING:
    from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping
    from typing import Optional, TypeVar

    from stereomolgraph.graphs import (
        AtomId,
        CondensedReactionGraph,
        MolGraph,
        StereoCondensedReactionGraph,
        StereoMolGraph,
    )
    from stereomolgraph.graphs.scrg import Change
    from stereomolgraph.stereodescriptors import Stereo

    KT = TypeVar("KT", bound=Hashable, covariant=True)
    VT = TypeVar("VT", bound=Hashable, covariant=True)

class _Parameters(NamedTuple):
    """
    Parameters of the algorithm.
    :ivar g1_nbrhd: Neighborhood of the first graph
    :ivar g2_nbrhd: Neighborhood of the second graph
    :ivar g1_labels: Labels of the first graph
    :ivar g2_labels: Labels of the second graph
    :ivar nodes_of_g1Labels: Nodes of the first graph grouped by labels
    :ivar nodes_of_g2Labels: Nodes of the second graph grouped by labels
    :ivar g1_degree: Degree of the first graph
    :ivar g2_nodes_of_degree: Nodes of the second graph grouped by degree
    :ivar g1_stereo: Stereochemistry of the first graph
    :ivar g2_stereo: Stereochemistry of the second graph
    :ivar g1_stereo_changes: Stereochemistry changes of the first graph
    :ivar g2_stereo_changes: Stereochemistry changes of the second graph
    """

    # Neighborhood
    g1_nbrhd: Mapping[AtomId, set[AtomId]]
    g2_nbrhd: Mapping[AtomId, set[AtomId]]
    # atomid: label
    g1_labels: Mapping[AtomId, int]
    g2_labels: Mapping[AtomId, int]
    # label: set of atomids
    nodes_of_g1Labels: Mapping[int, set[AtomId]]
    nodes_of_g2Labels: Mapping[int, set[AtomId]]
    # degree: set of atomids
    g1_degree: Mapping[AtomId, int]
    g2_nodes_of_degree: Mapping[int, set[AtomId]]
    # atomid: list of stereos containing the atom
    g1_stereo: Mapping[AtomId, list[Stereo]]
    g2_stereo: Mapping[AtomId, list[Stereo]]

    g1_stereo_changes: Mapping[AtomId, Mapping[Change, list[Stereo]]]
    g2_stereo_changes: Mapping[AtomId, Mapping[Change, list[Stereo]]]


class _State(NamedTuple):
    """
    State of the algorithm.
    :ivar mapping:
    :ivar inverted_mapping:
    :ivar frontier1: neighbors of mapped atoms in g1
    :ivar external1: atoms not in mapping and not in frontier1
    :ivar frontier2: neighbors of mapped atoms in g2
    :ivar external2: atoms not in mapping and not in frontier2
    """

    mapping: dict[int, int]
    inverted_mapping: dict[int, int]

    frontier1: set[int]  # neighbors of mapped atoms in g1
    external1: set[int]  # atoms not in mapping and not in frontier1

    frontier2: set[int]  # neighbors of mapped atoms in g2
    external2: set[int]  # atoms not in mapping and not in frontier2


def group_keys_by_value(many_to_one: dict[KT, VT]) -> dict[VT, set[KT]]:
    """Inverts a many-to-one mapping to create a one-to-many mapping.

    Converts a dictionary where multiple keys may point to the same value
    into a dictionary where each original value maps to a set of all original
    keys.

    >>> group_keys_by_value({"a": 1, "b": 1, "c": 2, "d": 3, "e": 3})
    {1: {'a', 'b'}, 2: {'c'}, 3: {'e', 'd'}}

    :param many_to_one: Dictionary to invert
    :return: Inverted dictionary where each value maps to a set of keys
    """
    inverted: defaultdict[VT, set[KT]] = defaultdict(set)
    for key, value in many_to_one.items():
        inverted[value].add(key)
    return dict(inverted)


def bfs_layers(
    neighbor_dict: Mapping[AtomId, Iterable[AtomId]],
    sources: Iterable[AtomId] | AtomId,
) -> Iterator[list[AtomId]]:
    """
    Generates layers of the graph starting from the source atoms.
    Each layer contains all nodes that are at the same distance from the
    sources.
    The first layer contains the sources.

    :param neighbor_dict: Dictionary of neighbors for each atom
    :param sources: Sources to start from
    """
    if isinstance(sources, int):
        sources = [sources]
    
    current_layer = list(sources)
    visited = set(sources)

    if any(source not in neighbor_dict for source in current_layer):
        raise ValueError("Source atom not in molecule")

    # this is basically BFS, except that the current layer only stores the
    # nodes at same distance from sources at each iteration
    while current_layer:
        yield current_layer
        next_layer: list[AtomId] = []
        for node in current_layer:
            for child in neighbor_dict[node]:
                if child not in visited:
                    visited.add(child)
                    next_layer.append(child)
        current_layer = next_layer


def _sanity_check_and_init(
    g1: StereoMolGraph | MolGraph,
    g2: StereoMolGraph | MolGraph,
    labels: Optional[Iterable[str]] = None,
    color_refine: bool | int = False,
    stereo: bool = False,
    stereo_change: bool = False,
    subgraph: bool = False,
) -> Optional[tuple[_Parameters, _State]]:
    if stereo_change and not stereo:
        raise ValueError("Stereo change is only available for stereo graphs.")
    if subgraph and color_refine:
        raise ValueError(
            "Subgraph isomorphism is not compatible with color refinement."
        )

    g1_nbrhd = g1.neighbors
    g2_nbrhd = g2.neighbors

    if len(g1_nbrhd) == 0 or len(g2_nbrhd) == 0:
        return None

    elif not subgraph:
        if len(g1_nbrhd) != len(g2_nbrhd):
            return None

        if sorted(len(n) for n in g1_nbrhd.values()) != sorted(
            len(n) for n in g2_nbrhd.values()
        ):
            return None

    elif subgraph:
        if len(g1_nbrhd) > len(g2_nbrhd):
            ValueError("The second graph must be larger than the first one.")
        elif len(g1_nbrhd) == len(g2_nbrhd):
            raise ValueError(
                "Both graphs have the same number of atoms. "
                "Do not use subgraph isomorphism in this case."
            )
        elif Counter(len(nbr) for nbr in g1_nbrhd.values()) > Counter(
            len(nbr) for nbr in g2_nbrhd.values()
        ):
            return None

    atom_labels = ("atom_type", *labels) if labels else ("atom_type",)

    max_iter = color_refine
    
    if color_refine: # and not stereo and not stereo_change:
        g1_labels = color_refine_mg(
            g1, atom_labels=atom_labels, max_iter=max_iter
        )
        g2_labels = color_refine_mg(
            g2, atom_labels=atom_labels, max_iter=max_iter
        )

    #elif color_refine and stereo and not stereo_change:
    #
        # TODO: Implement color refinement for stereo graphs
        #g1_labels = color_refine_smg(
        #    g1, atom_labels=atom_labels, max_iter=max_iter
        #)
        #g2_labels = color_refine_smg(
        #    g2, atom_labels=atom_labels, max_iter=max_iter
        #)

    #elif color_refine and stereo and stereo_change:
    #
        #g1_labels = color_refine_scrg(
        #    g1, atom_labels=atom_labels, max_iter=max_iter
        #)
        #g2_labels = color_refine_scrg(
        ##    g2, atom_labels=atom_labels, max_iter=max_iter
        #)

    elif not color_refine:
        g1_labels = label_hash(g1, atom_labels=atom_labels)
        g2_labels = label_hash(g2, atom_labels=atom_labels)

    else:
        raise ValueError("Invalid combination of parameters.")
    
    
    g1_labels_counter = Counter(g1_labels.values())
    g2_labels_counter = Counter(g2_labels.values())

    if not subgraph and g1_labels_counter != g2_labels_counter:
        return None

    elif subgraph and g1_labels_counter > g2_labels_counter:
        return None

    g1_stereo: defaultdict[AtomId, list[Stereo]] = defaultdict(list)
    g2_stereo: defaultdict[AtomId, list[Stereo]] = defaultdict(list)

    if stereo:
        if TYPE_CHECKING:
            assert isinstance(g1, StereoMolGraph)
            assert isinstance(g2, StereoMolGraph)

        for s in g1.stereo.values():
            for atom in s.atoms:
                g1_stereo[atom].append(s)

        for s in g2.stereo.values():
            for atom in s.atoms:
                g2_stereo[atom].append(s)

    g1_stereo_changes: defaultdict[AtomId, defaultdict[Change, list[Stereo]]] = defaultdict(lambda: defaultdict(list)) # type: ignore

    g2_stereo_changes: defaultdict[AtomId, defaultdict[Change, list[Stereo]]] = defaultdict(lambda: defaultdict(list)) # type: ignore

    if stereo_change:
        if TYPE_CHECKING:
            assert isinstance(g1, StereoCondensedReactionGraph)
            assert isinstance(g2, StereoCondensedReactionGraph)
        
        for _, stereo_change_dict in g1.atom_stereo_changes.items():
            for stereo_change_enum, atom_stereo in stereo_change_dict.items():
                if atom_stereo is not None:
                    for atom in atom_stereo.atoms:
                        g1_stereo_changes[atom][stereo_change_enum].append(
                            atom_stereo
                        )

        for _, stereo_change_dict in g2.atom_stereo_changes.items():
            for stereo_change_enum, atom_stereo in stereo_change_dict.items():
                if atom_stereo is not None:
                    for atom in atom_stereo.atoms:
                        g2_stereo_changes[atom][stereo_change_enum].append(
                            atom_stereo
                        )

        for _, stereo_change_dict in g1.bond_stereo_changes.items():
            for stereo_change_enum, bond_stereo in stereo_change_dict.items():
                if bond_stereo is not None:
                    for atom in bond_stereo.atoms:
                        g1_stereo_changes[atom][stereo_change_enum].append(
                            bond_stereo
                        )

        for _, stereo_change_dict in g2.bond_stereo_changes.items():
            for stereo_change_enum, bond_stereo in stereo_change_dict.items():
                if bond_stereo is not None:
                    for atom in bond_stereo.atoms:
                        g2_stereo_changes[atom][stereo_change_enum].append(
                            bond_stereo
                        )

    g1_degree = {a: len(n_set) for a, n_set in g1_nbrhd.items()}
    g2_degree = {a: len(n_set) for a, n_set in g2_nbrhd.items()}

    params = _Parameters(
        g1_nbrhd,
        g2_nbrhd,
        g1_labels,
        g2_labels,
        group_keys_by_value(g1_labels),
        group_keys_by_value(g2_labels),
        g1_degree,
        group_keys_by_value(g2_degree),
        g1_stereo,
        g2_stereo,
        g1_stereo_changes,
        g2_stereo_changes,
    )

    state = _State({}, {}, set(), set(g1_nbrhd), set(), set(g2_nbrhd))

    return params, state


def _wrap_all(
    *funcs: Callable[[AtomId, AtomId, _State, _Parameters], bool],
) -> Callable[[AtomId, AtomId, _State, _Parameters], bool]:
    def wrapper(a: AtomId, b: AtomId, state: _State, params: _Parameters) -> bool:
        return all(f(a, b, state, params) for f in funcs)
    
    return wrapper


def vf2pp_all_isomorphisms(
    g1: MolGraph
    | StereoMolGraph
    | CondensedReactionGraph
    | StereoCondensedReactionGraph,
    g2: MolGraph
    | StereoMolGraph
    | CondensedReactionGraph
    | StereoCondensedReactionGraph,
    labels: Optional[Iterable[str]] = None,
    bond_labels: Optional[Iterable[str]] = None,
    color_refine: bool | int = True,
    stereo: bool = False,
    stereo_change: bool = False,
    subgraph: bool = False,
) -> Iterator[dict[AtomId, AtomId]]:
    if params_state := _sanity_check_and_init(
        g1, g2, labels, color_refine, stereo, stereo_change, subgraph
    ):
        params, state = params_state
    else:
        return  # if no isomorphisms return like an empty generator

    # setup helper function based on input parameters

    feasibility_funcs: list[Callable[[AtomId, AtomId, _State, _Parameters], bool]] = []
    if subgraph:
        feasibility_funcs.append(_subgraph_feasibility)
        if stereo:
            feasibility_funcs.append(_subgraph_stereo_feasibility)
        if stereo_change:
            feasibility_funcs.append(_subgraph_stereo_change_feasibility)
            
    elif not subgraph:
        feasibility_funcs.append(_graph_feasibility)
        if stereo:
            feasibility_funcs.append(_stereo_feasibility)
        if stereo_change:
            feasibility_funcs.append(_stereo_change_feasibility)
    else:
        raise ValueError("Invalid combination of parameters.")

    feasibility = _wrap_all(*feasibility_funcs) # type: ignore
    revert_state = _revert_state
    update_state = _update_state
    find_candidates = (_find_subgraph_candidates if subgraph
                       else _find_candidates)

    # to avoid overhead
    mapping = state.mapping
    inverted_mapping = state.inverted_mapping
    termination_length = len(g1)

    # Initialize the stack
    node_order: list[AtomId] = _matching_order(params)
    candidates: set[AtomId] = find_candidates(node_order[0], state, params)

    stack: list[tuple[AtomId, set[AtomId]]] = []
    stack.append((node_order[0], candidates))

    # Index of the node from the order, currently being examined
    matching_atom_index = 1

    while stack:
        matching_atom, candidates = stack[-1]
        candidate = candidates.pop() if candidates else None
        if candidate is None:
            # If no remaining candidates, return to a previous state,
            # and follow another branch
            stack.pop()
            matching_atom_index -= 1
            if stack:
                # Pop the previously added u-v pair, and look for
                # a different candidate _v for u
                last_atom1, _ = stack[-1]
                last_atom2 = mapping.pop(last_atom1)
                inverted_mapping.pop(last_atom2)
                revert_state(last_atom1, last_atom2, state, params)
            continue

        mapping[matching_atom] = candidate
        inverted_mapping[candidate] = matching_atom

        if feasibility(matching_atom, candidate, state, params):
            if len(mapping) == termination_length:
                yield mapping.copy()
                mapping.pop(matching_atom)
                inverted_mapping.pop(candidate)
                continue

            update_state(matching_atom, candidate, state, params)
            # Append the next node and its candidates to the stack
            matching_atom = node_order[matching_atom_index]
            candidates = find_candidates(matching_atom, state, params)
            stack.append((matching_atom, candidates))
            matching_atom_index += 1

        else:  # if not feaseble
            mapping.pop(matching_atom)
            inverted_mapping.pop(candidate)


def _graph_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
):
    g1_nbrhd, g2_nbrhd, g1_labels, g2_labels, *_ = params # type: ignore
    _, _, frontier1, external1, frontier2, external2 = state

    t1_labels: list[int] = []
    t2_labels: list[int] = []
    t1_tilde_labels: list[int] = []
    t2_tilde_labels: list[int] = []

    for n in g1_nbrhd[u]:
        if n in external1:
            t1_tilde_labels.append(g1_labels[n])
        elif n in frontier1:
            t1_labels.append(g1_labels[n])

    for n in g2_nbrhd[v]:
        if n in external2:
            t2_tilde_labels.append(g2_labels[n])
        elif n in frontier2:
            t2_labels.append(g2_labels[n])

    t1_labels.sort()
    t2_labels.sort()

    if t1_labels != t2_labels:
        return False

    t1_tilde_labels.sort()
    t2_tilde_labels.sort()

    if t1_tilde_labels != t2_tilde_labels:
        return False

    return True

def _subgraph_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool:
    g1_nbrhd, g2_nbrhd, g1_labels, g2_labels, *_ = params # type: ignore
    _, _, frontier1, external1, frontier2, external2 = state

    counter1 = Counter(g1_labels[n] for n in g1_nbrhd[u] if n in frontier1)
    counter2 = Counter(g2_labels[n] for n in g2_nbrhd[v] if n in frontier2)
    if counter1 > counter2:
        return False

    counter1 = Counter(g1_labels[n] for n in g1_nbrhd[u] if n in external1)
    counter2 = Counter(g2_labels[n] for n in g2_nbrhd[v] if n in external2)
    return counter1 <= counter2

def _stereo_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool:
    s1 = [
        stereo.__class__(
            atoms=tuple([state.mapping[a] for a in stereo.atoms]),
            parity=stereo.parity,
        )
        for stereo in params.g1_stereo[u]
        if all([a in state.mapping for a in stereo.atoms])
    ]

    s2 = [
        stereo
        for stereo in params.g2_stereo[v]
        if all([a in state.inverted_mapping for a in stereo.atoms])
    ]

    if len(s2) != len(s1):
        return False

    if all(s in s2 for s in s1):
        return True
    return False

def _subgraph_stereo_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool: ...

def _stereo_change_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool:
    s1 = {
        (
            stereo_change,
            stereo.__class__(
                atoms=tuple([state.mapping[a] for a in stereo.atoms]),
                parity=stereo.parity,
            ),
        )
        for stereo_change, stereo_list in params.g1_stereo_changes[u].items()
        for stereo in stereo_list
        if stereo is not None # type: ignore
        and all([a in state.mapping for a in stereo.atoms])
    }

    s2 = {
        (stereo_change, stereo)
        for stereo_change, stereo_list in params.g2_stereo_changes[u].items()
        for stereo in stereo_list
        if stereo is not None # type: ignore
        and all([a in state.inverted_mapping for a in stereo.atoms])
    }

    if s1 == s2:
        return True
    return False

def _subgraph_stereo_change_feasibility(
    u: AtomId, v: AtomId, state: _State, params: _Parameters
) -> bool: ...
    # TODO: Implement subgraph stereo change feasibility check

def _matching_order(params: _Parameters) -> list[AtomId]:
    g1_nbrhd, _, g1_labels, _, _, nodes_of_g2Labels, *_ = params

    V1_unordered = set(g1_labels.keys())
    label_rarity = {
        label: len(nodes) for label, nodes in nodes_of_g2Labels.items()
    }
    used_degrees = {node: 0 for node in g1_nbrhd}
    node_order: list[int] = []

    while V1_unordered:
        max_rarity = min(label_rarity[g1_labels[x]] for x in V1_unordered)
        rarest_nodes = [
            n for n in V1_unordered if label_rarity[g1_labels[n]] == max_rarity
        ]
        max_node: AtomId = max(rarest_nodes, key={a: len(n_set) for a, n_set in g1_nbrhd.items()}.get,) # type: ignore
        assert isinstance(max_node, int)
        for dlevel_nodes in bfs_layers(g1_nbrhd, max_node):
            nodes_to_add = dlevel_nodes.copy()
            while nodes_to_add:
                max_used_degree = max(used_degrees[n] for n in nodes_to_add)
                max_used_degree_nodes = [
                    n
                    for n in nodes_to_add
                    if used_degrees[n] == max_used_degree
                ]
                max_degree = max(
                    len(g1_nbrhd[n]) for n in max_used_degree_nodes
                )
                max_degree_nodes = [
                    n
                    for n in max_used_degree_nodes
                    if len(g1_nbrhd[n]) == max_degree
                ]
                next_node = min(
                    max_degree_nodes, key=lambda x: label_rarity[g1_labels[x]]
                )

                node_order.append(next_node)
                for node in g1_nbrhd[next_node]:
                    used_degrees[node] += 1

                nodes_to_add.remove(next_node)
                label_rarity[g1_labels[next_node]] -= 1
                V1_unordered.discard(next_node)

    return node_order


def _find_candidates(
    u: AtomId, state: _State, params: _Parameters
) -> set[AtomId]:
    (
        g1_nbrhd,
        g2_nbrhd,
        g1_labels,
        _,
        _,
        nodes_of_g2Labels,
        g1_deg,
        g2_a_of_deg,
        *_,
    ) = params
    mapping, inverted_mapping, *_, external2 = state

    covered_nbrs = [nbr for nbr in g1_nbrhd[u] if nbr in mapping]

    if not covered_nbrs:
        candidates = set(nodes_of_g2Labels[g1_labels[u]])
        candidates.intersection_update(external2, g2_a_of_deg[g1_deg[u]])
        candidates.difference_update(inverted_mapping)

        return candidates

    nbr1 = covered_nbrs[0]
    candidates = set(g2_nbrhd[mapping[nbr1]])

    for nbr1 in covered_nbrs[1:]:
        candidates.intersection_update(g2_nbrhd[mapping[nbr1]])

    candidates.difference_update(inverted_mapping)
    candidates.intersection_update(
        nodes_of_g2Labels[g1_labels[u]], g2_a_of_deg[g1_deg[u]]
    )
    return candidates


def _find_subgraph_candidates(
    u: AtomId, state: _State, params: _Parameters
) -> set[AtomId]:
    (
        g1_nbrhd,
        g2_nbrhd,
        g1_labels,
        _,
        _,
        nodes_of_g2Labels,
        *_,
    ) = params
    mapping, inverted_mapping, *_, external2 = state

    covered_nbrs = [nbr for nbr in g1_nbrhd[u] if nbr in mapping]

    if not covered_nbrs:
        candidates = set(nodes_of_g2Labels[g1_labels[u]])
        candidates.intersection_update(external2)
        candidates.difference_update(inverted_mapping)
        # candidates.intersection_update(g2_a_of_deg[degree] for degree
        #                               in g2_a_of_deg if degree < g1_deg[u])
        return candidates

    nbr1 = covered_nbrs[0]
    common_nodes = set(g2_nbrhd[mapping[nbr1]])

    for nbr1 in covered_nbrs[1:]:
        common_nodes.intersection_update(g2_nbrhd[mapping[nbr1]])

    common_nodes.difference_update(inverted_mapping)
    common_nodes.intersection_update(nodes_of_g2Labels[g1_labels[u]])

    # common_nodes.intersection_update(g2_a_of_deg[degree] for degree
    #                                 in g2_a_of_deg if degree < g1_deg[u])

    return common_nodes


def _update_state(
    new_atom1: AtomId, new_atom2: AtomId, state: _State, params: _Parameters
) -> None:
    g1_nbrhd, g2_nbrhd, *_ = params
    mapping, inverted_mapping, frontier1, external1, frontier2, external2 = (
        state
    )

    unmapped_neighbors1 = {n for n in g1_nbrhd[new_atom1] if n not in mapping}
    frontier1 |= unmapped_neighbors1
    external1 -= unmapped_neighbors1
    frontier1.discard(new_atom1)
    external1.discard(new_atom1)

    unmapped_neighbors2 = {
        n for n in g2_nbrhd[new_atom2] if n not in inverted_mapping
    }
    frontier2 |= unmapped_neighbors2
    external2 -= unmapped_neighbors2
    frontier2.discard(new_atom2)
    external2.discard(new_atom2)

    return


def _revert_state(
    last_atom1: AtomId, last_atom2: AtomId, state: _State, params: _Parameters
) -> None:
    # If the node we want to remove from the mapping, has at least one
    # covered neighbor, add it to frontier1.
    g1_nbrhd, g2_nbrhd, *_ = params
    mapping, inverted_mapping, frontier1, external1, frontier2, external2 = (
        state
    )

    has_covered_neighbor = False
    for neighbor in g1_nbrhd[last_atom1]:
        if neighbor in mapping:
            # if a neighbor of the excluded node1 is in the mapping,
            # keep node1 in frontier1
            has_covered_neighbor = True
            frontier1.add(last_atom1)
        else:
            # check if its neighbor has another connection with a covered node.
            # If not, only then exclude it from frontier1
            if any(nbr in mapping for nbr in g1_nbrhd[neighbor]):
                continue
            frontier1.discard(neighbor)
            external1.add(neighbor)

    # Case where the node is not present in neither the mapping nor frontier1.
    # By definition, it should belong to external1
    if not has_covered_neighbor:
        external1.add(last_atom1)

    has_covered_neighbor = False
    for neighbor in g2_nbrhd[last_atom2]:
        if neighbor in inverted_mapping:
            has_covered_neighbor = True
            frontier2.add(last_atom2)
        else:
            if any(nbr in inverted_mapping for nbr in g2_nbrhd[neighbor]):
                continue
            frontier2.discard(neighbor)
            external2.add(neighbor)

    if not has_covered_neighbor:
        external2.add(last_atom2)




from __future__ import annotations

from collections import Counter
from copy import deepcopy
from types import MappingProxyType
from typing import TYPE_CHECKING, Literal, TypeVar

import numpy as np

from stereomolgraph.algorithms.color_refine import color_refine_mg
from stereomolgraph.algorithms.isomorphism import vf2pp_all_isomorphisms
from stereomolgraph.coords import are_planar
from stereomolgraph.graph2rdmol import stereo_mol_graph_to_rdmol
from stereomolgraph.graphs.mg import AtomId, Bond, MolGraph
from stereomolgraph.rdmol2graph import stereo_mol_graph_from_rdmol
from stereomolgraph.stereodescriptors import (
    AtomStereo,
    BondStereo,
    PlanarBond,
    Tetrahedral,
    TrigonalBipyramidal,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping
    from typing import Self

    from rdkit import Chem

    from stereomolgraph.coords import Geometry
    A = TypeVar("A", bound=tuple[int, ...], covariant=True)
    P = TypeVar("P", bound=None | Literal[1, 0, -1], covariant=True)


class StereoMolGraph(MolGraph):
    """
    :class:`MolGraph` with the ability to store stereochemistry information
    for atoms and bonds.

    Two graphs compare equal, if they are isomorphic and have the same
    stereochemistry.
    """
    __slots__ = ("_atom_stereo", "_bond_stereo")

    _atom_stereo: dict[int, AtomStereo]
    _bond_stereo: dict[Bond, BondStereo]

    def __init__(self, mol_graph: None | MolGraph = None):
        super().__init__(mol_graph)
        if mol_graph and isinstance(mol_graph, StereoMolGraph):
            self._atom_stereo = deepcopy(mol_graph._atom_stereo)
            self._bond_stereo = deepcopy(mol_graph._bond_stereo)
        else:
            self._atom_stereo = {}
            self._bond_stereo = {}

    def __hash__(self) -> int:
        color_dict: dict[AtomId, int] = color_refine_mg(self, )
        connectivity_hash = set(Counter(color_dict.values()).items())
        atom_stereo = {s.__class__( tuple([color_dict[a] for a in s.atoms]),
                                   s.parity)
                                   for s in self.atom_stereo.values()}
        bond_stereo = {s.__class__(tuple([color_dict[a] for a in s.atoms]),
                                   s.parity)
                                   for s in self.bond_stereo.values()}
        return hash((frozenset(connectivity_hash),
                     frozenset(atom_stereo),
                     frozenset(bond_stereo)))

    @property
    def stereo(self) -> Mapping[AtomId | Bond, AtomStereo | BondStereo]:
        return MappingProxyType(self._atom_stereo | self._bond_stereo)

    @property
    def atom_stereo(self) -> Mapping[AtomId, AtomStereo]:
        return MappingProxyType(self._atom_stereo)

    @property
    def bond_stereo(self) -> Mapping[Bond, BondStereo]:
        return MappingProxyType(self._bond_stereo)

    def get_atom_stereo(
        self, atom: AtomId
    ) -> None | AtomStereo:
        """Returns the stereo information of the atom if it exists else None.
        Raises a ValueError if the atom is not in the graph.

        :param atom: atom
        :param default: Default value if no stereo information is found,
                        defaults to None
        :return: Stereo information of atom
        """
        if atom in self._atom_attrs:
            if s := self._atom_stereo.get(atom, None):
                return s
            else:
                return None
                #return NoStereo(atoms=(atom, *list(self.bonded_to(atom))))
        else:
            raise ValueError(f"Atom {atom} is not in the graph")

    def set_atom_stereo(self, atom_stereo: AtomStereo):
        """Adds stereo information to the graph

        :param atom: Atoms to be used for chiral information
        :param stereo: Chiral information
        """
        atom = atom_stereo.central_atom
        if atom in self._atom_attrs:
            assert atom in atom_stereo.atoms
            self._atom_stereo[atom] = atom_stereo
        else:
            raise ValueError(f"Atom {atom} is not in the graph")

    def delete_atom_stereo(self, atom: AtomId):
        """Deletes stereo information from the graph

        :param atom: Atom to be used for stereo information
        """
        del self._atom_stereo[atom]

    def get_bond_stereo(
        self, bond: Iterable[int]
    ) -> None | BondStereo:
        """Gets the stereo information of the bond or None
        if it does not exist.
        Raises a ValueError if the bond s not in the graph.

        :param bond: Bond
        :return: stereo information of bond
        """
        bond = Bond(bond)
        bond_stereo = self._bond_stereo.get(Bond(bond), None)
        if bond_stereo:
            return bond_stereo
        elif bond in self._bond_attrs:
            return None
        else:
            raise ValueError(f"Bond {bond} is not in the graph")

    def set_bond_stereo(
        self, bond_stereo: BondStereo
    ):
        """Stets the stereo information of the bond

        :param bond: Bond
        :param bond_stereo: Stereo information of the bond
        """

        bond = Bond(bond_stereo.bond)
        if bond in self._bond_attrs:
            self._bond_stereo[bond] = bond_stereo
        else:
            raise ValueError(f"Bond {bond} is not in the graph")

    def delete_bond_stereo(self, bond: Iterable[int]):
        """Deletes the stereo information of the bond

        :param bond: Bond
        """
        del self._bond_stereo[Bond(bond)]

    def remove_atom(self, atom: int):
        """Removes an atom from the graph and deletes all chiral information
        associated with it

        :param atom: Atom
        """
        for a, atom_stereo in self._atom_stereo.copy().items():
            if atom in atom_stereo.atoms:
                self.delete_atom_stereo(a)

        for bond, bond_stereo in self._bond_stereo.copy().items():
            if atom in bond_stereo.atoms:
                self.delete_bond_stereo(bond)
        super().remove_atom(atom)

    def copy(self) -> Self:
        """
        :return: returns a copy of self
        """
        new_graph = super().copy()
        new_graph._atom_stereo = deepcopy(self._atom_stereo)
        new_graph._bond_stereo = deepcopy(self._bond_stereo)
        return new_graph

    def relabel_atoms(
        self, mapping: dict[int, int], copy: bool = True
    ) -> Self:
        """
        Relabels the atoms of the graph and the chiral information accordingly

        :param mapping: Mapping of old atom ids to new atom ids
        :param copy: If the graph should be copied before relabeling,
                     defaults to True
        :return: Returns the relabeled graph
        """
        new_atom_stereo_dict = self._atom_stereo.__class__()
        new_bond_stereo_dict = self._bond_stereo.__class__()

        for central_atom, stereo in self._atom_stereo.items():
            new_central_atom = mapping.get(central_atom, central_atom)
            new_atom_stereo_atoms = tuple(
                mapping.get(atom, atom) for atom in stereo.atoms
            )
            new_atom_stereo = stereo.__class__(
                new_atom_stereo_atoms, stereo.parity
            )
            new_atom_stereo_dict[new_central_atom] = new_atom_stereo

        for bond, bond_stereo in self._bond_stereo.items():
            new_bond = tuple(mapping.get(atom, atom) for atom in bond)
            new_bond_stereo_atoms = tuple(
                mapping.get(atom, atom) for atom in bond_stereo.atoms
            )
            new_bond_stereo = bond_stereo.__class__(
                new_bond_stereo_atoms, bond_stereo.parity
            )
            new_bond_stereo_dict[frozenset(new_bond)] = new_bond_stereo

        if copy is True:
            graph = super().relabel_atoms(mapping, copy=True)
            graph._atom_stereo = new_atom_stereo_dict
            graph._bond_stereo = new_bond_stereo_dict
            return graph

        elif copy is False:
            super().relabel_atoms(mapping, copy=False)
            self._atom_stereo = new_atom_stereo_dict
            self._bond_stereo = new_bond_stereo_dict
            return self

    def subgraph(self, atoms: Iterable[int]) -> Self:
        """Returns a subgraph of the graph with the given atoms and the chiral
        information accordingly

        :param atoms: Atoms to be used for the subgraph
        :return: Subgraph
        """
        new_graph = super().subgraph(atoms)

        for central_atom, atoms_atom_stereo in self._atom_stereo.items():
            atoms_set = set((*atoms_atom_stereo.atoms, central_atom))
            if all(atom in atoms for atom in atoms_set):
                new_graph.set_atom_stereo(atoms_atom_stereo)

        for _bond, bond_stereo in self._bond_stereo.items():
            if all(atom in atoms for atom in bond_stereo.atoms):
                new_graph.set_bond_stereo(bond_stereo)
        return new_graph

    def enantiomer(self) -> Self:
        """
        Creates the enantiomer of the StereoMolGraph by inversion of all atom
        stereocenters. The result can be identical to the molecule itself if
        no enantiomer exists.

        :return: Enantiomer
        """
        enantiomer = self.copy()
        for atom in self.atoms:
            if stereo := self.get_atom_stereo(atom):
                enantiomer.set_atom_stereo(stereo.invert())
        return enantiomer

    def _to_rdmol(
        self,
        generate_bond_orders:bool=False,
        allow_charged_fragments:bool=False,
        charge:int=0
    ) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
        """
        Creates a RDKit mol object using the connectivity of the mol graph.
        Stereochemistry is added to the mol object.

        :return: RDKit molecule
        """
        return stereo_mol_graph_to_rdmol(self,
                                          generate_bond_orders=generate_bond_orders,
                                          allow_charged_fragments=allow_charged_fragments,
                                          charge=charge)
        

    @classmethod
    def from_rdmol(cls, rdmol:Chem.Mol, use_atom_map_number:bool=False
                   ) -> Self:
        """
        Creates a StereoMolGraph from an RDKit Mol object.
        All hydrogens have to be explicit.
        Stereo information is conserved for tetrahedral atoms and
        double bonds.

        :param rdmol: RDKit Mol object
        :param use_atom_map_number: If the atom map number should be used
                                    instead of the atom index, Default: False
        :return: StereoMolGraph
        """
        smg = stereo_mol_graph_from_rdmol(cls, rdmol,
                                          use_atom_map_number=use_atom_map_number)
        assert isinstance(smg, cls), ("StereoMolGraph.from_rdmol did not"
                                      " return a StereoMolGraph")
        return smg

    def _set_atom_stereo_from_geometry(self, geo: Geometry):
        for atom in range(geo.n_atoms):
            first_neighbors = list(self.bonded_to(atom))

            # extends the first layer of neighbors to the second layer
            # (if planar)
            # needed to find double bonds
            if len(first_neighbors) < 3:
                pass
            elif (len(first_neighbors) == 3
                  and are_planar(geo.coords[[atom, *first_neighbors]])):
                for neighbor in first_neighbors:
                    next_layer = set(self.bonded_to(neighbor))

                    for i in first_neighbors:
                        next_layer.add(i)
                    next_layer -= set((neighbor, atom))

                    if len(next_layer) != 4:
                        continue

                    elif are_planar(geo.coords.take(tuple(next_layer),
                                                    axis=0)):
                        bonded_to_atom = (
                            outer_atom
                            for outer_atom in next_layer
                            if self.has_bond(atom, outer_atom)
                        )
                        bonded_to_neighbor = (
                            outer_atom
                            for outer_atom in next_layer
                            if self.has_bond(neighbor, outer_atom)
                        )
                        atom_ids = (
                            *bonded_to_atom,
                            atom,
                            neighbor,
                            *bonded_to_neighbor,
                        )
                        assert len(atom_ids) == 6
                        double_bond = PlanarBond.from_coords(
                            atom_ids, geo.coords.take(atom_ids, axis=0))
                        
                        self.set_bond_stereo(double_bond)

            elif len(first_neighbors) == 3:
                pass
            else:
                first_neighbors_coords = geo.coords[first_neighbors]
                if are_planar(first_neighbors_coords):
                    pass

                elif len(first_neighbors) == 4:
                    stereo_atoms = (atom, *first_neighbors)
                    assert len(stereo_atoms) == 5
                    stereo_coords = geo.coords.take(stereo_atoms, axis=0)
                    atoms_atom_stereo = Tetrahedral.from_coords(stereo_atoms,
                                                                stereo_coords)

                    self.set_atom_stereo(atoms_atom_stereo)

                elif len(first_neighbors) == 5:
                    stereo_atoms = (atom, *first_neighbors)
                    assert len(stereo_atoms) == 6
                    stereo_coords = geo.coords.take(stereo_atoms, axis=0)
                    atoms_atom_stereo = TrigonalBipyramidal.from_coords(
                        stereo_atoms, stereo_coords)
                    self.set_atom_stereo(atoms_atom_stereo)


    @classmethod
    def compose(cls, mol_graphs: Iterable[MolGraph]) -> Self:
        """Creates a MolGraph object from a list of MolGraph objects.
        
        Duplicate nodes or edges are overwritten, such that the resulting
        graph only contains one node or edge with that name. Duplicate
        attributes of duplicate nodes, edges and the stereochemistry are also
        overwritten in order of iteration.

        :param mol_graphs: list of MolGraph objects
        :return: Returns MolGraph
        """

        graph = cls(super().compose(mol_graphs))
        for mol_graph in mol_graphs:
            graph._atom_stereo.update(cls(mol_graph)._atom_stereo)
            graph._bond_stereo.update(cls(mol_graph)._bond_stereo)
        return graph

    @classmethod
    def from_geometry_and_bond_order_matrix(
        cls: type[Self],
        geo: Geometry,
        matrix: np.ndarray,
        threshold: float = 0.5,
        include_bond_order: bool = False,
    ) -> Self:
        """
        Creates a CiralMolGraph object from a Geometry and a bond order matrix

        :param geo: Geometry
        :param matrix: Bond order matrix
        :param threshold: Threshold for bonds to be included as edges,
                          defaults to 0.5
        :param include_bond_order: If bond orders should be included as edge
                                    attributes, defaults to False
        :return: Returns MolGraph
        """
        mol_graph = super().from_geometry_and_bond_order_matrix(
            geo,
            matrix=matrix,
            threshold=threshold,
            include_bond_order=include_bond_order,
        )
        graph = cls(mol_graph)
        graph._set_atom_stereo_from_geometry(geo)
        return graph

    def get_isomorphic_mappings(
        self, other: MolGraph, stereo:bool=True
    ) -> Iterator[dict[int, int]]:
        """Isomorphic mappings between "self" and "other".

        Generates all isomorphic mappings between "other" and "self".
        All atoms and bonds have to be present in both graphs.
        The Stereochemistry is preserved in the mappings.

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        :raises TypeError: Not defined for objects different types
        """

        return vf2pp_all_isomorphisms(
            self,
            other,
            color_refine=True, #TODO: implement color refinement
            stereo=stereo,
            stereo_change=False,
            subgraph=False,
        )
        

    def get_subgraph_isomorphic_mappings(
        self, other: MolGraph, stereo: bool = True
    ) -> Iterator[dict[int, int]]:
        """Subgraph isomorphic mappings from "self" onto "other".
        Other can be of equal size or larger than "self".
        Generates all node-iduced subgraph isomorphic mappings.
        All atoms of "self" have to be present in "other".
        The bonds of "self" have to be the subset of the bonds of "other"
        relating to the nodes of "self".
        The Stereochemistry is preserved in the mappings.

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        :raises TypeError: Not defined for objects different types
        """
        if other.__class__ is not self.__class__:
            return
            yield
        
        return vf2pp_all_isomorphisms(
            self,
            other,
            color_refine=False,
            stereo=stereo,
            stereo_change=False,
            subgraph=True,
        )

    def is_stereo_valid(self) -> bool:
        """
        Checks if the bonds required to have the defined stereochemistry
        are present in the graph.

        :return: True if the stereochemistry is valid
        """
        for atom, stereo in self._atom_stereo.items():
            for neighbor in stereo.atoms[1:]:
                if not self.has_bond(atom, neighbor):
                    return False
        for bond, stereo in self._bond_stereo.items():
            if not self.has_bond(*bond):
                return False
            if {stereo.atoms[2], stereo.atoms[3]} != set(bond):
                return False
            if not self.has_bond(stereo.atoms[0], stereo.atoms[2]):
                return False
            if not self.has_bond(stereo.atoms[1], stereo.atoms[2]):
                return False
            if not self.has_bond(stereo.atoms[4], stereo.atoms[3]):
                return False
            if not self.has_bond(stereo.atoms[5], stereo.atoms[3]):
                return False
        return True




from __future__ import annotations

import itertools
from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Literal, Optional, TypeVar

    from stereomolgraph.graphs.mg import (
        AtomId,
        MolGraph,
    )

    N = TypeVar("N", bound=int)

def numpy_int_tuple_hash(
    arr: np.ndarray[tuple[int, ...], np.dtype[np.int64]],
    out: Optional[np.ndarray[tuple[Literal[1], ...], np.dtype[np.int64]]] = None,
) -> np.ndarray:
    """
    Mimics the python SipHash hashing function for tuples of integers
    with numpy int64 arrays.

    def SipHash(arr_slice):
        h = 0x345678
        mult = 1000003
        length = len(arr_slice)
        for i in range(1, length + 1):
            h = (h ^ arr_slice[i-1]) * mult
            mult += 82520 + 2 * (length - i)
        return h + 97531
    """
    # overflow is an expected behavior in this case
    with np.errstate(over="ignore"):
        arr_shape = arr.shape
        length = arr_shape[-1]
        if out is None:
            output = np.full(arr_shape[:-1], 0x345678, dtype=np.int64)
        else:
            output = out
            output.fill(0x345678)

        n = 82518 + 2 * length
        m = range(n, n - 2 * (length - 1), -2)
        mults = itertools.accumulate(m, initial=1000003)

        for idx, mult in enumerate(mults):
            output ^= arr[..., idx]
            output *= mult

        output += 97531
        return output

def numpy_int_set_hash(): ...

def label_hash(
    mg: MolGraph,
    atom_labels: Optional[Iterable[str]] = ("atom_type",),
    bond_labels: Optional[Iterable[str]] = None,
) -> dict[AtomId, int]:
    if atom_labels == ("atom_type",) and bond_labels is None:
        atom_hash = {
            atom: hash(mg.get_atom_type(atom))
            for atom in mg.atoms
        }

    elif atom_labels is None and bond_labels is None:
        atom_hash = {atom: 0 for atom in mg.atoms}

    elif atom_labels:
        atom_labels = sorted(atom_labels)
        atom_labels.append("atom_type")
        bond_labels = sorted(bond_labels) if bond_labels else []
        bond_labels.append("reaction")
        atom_hash = {atom:
              hash((
            tuple([(atom_label, label_dict.get(atom_label, None))
                   for atom_label in atom_labels]),
            tuple(sorted([(tuple(sorted(
                mg.get_bond_attributes(atom, nbr, bond_labels).items()))
                           for nbr in mg.bonded_to(atom))])),
            ))
              for atom, label_dict in mg.atoms_with_attributes.items()
        }
    else:
        raise ValueError("Invalid combination of atom and bond labels.")
    return atom_hash

def color_refine_mg(
    mg: MolGraph,
    max_iter: Optional[int] = None,
    atom_labels: Optional[Iterable[str]] = ("atom_type",),
    bond_labels: Optional[Iterable[str]] = None,
) -> dict[AtomId, int]:
    atom_label_hash = label_hash(mg, atom_labels, bond_labels)

    atom_hash: np.ndarray = np.array(
        [atom_label_hash[atom] for atom in mg.atoms], dtype=np.int64
    )

    n_atoms = np.int64(mg.n_atoms)
    id_arr = {atom: a_id for a_id, atom in enumerate(mg.atoms)}
    d = {
        id_arr[atom]: {id_arr[nbr] for nbr in mg.bonded_to(atom)}
        for atom in mg.atoms
    }

    grouped: defaultdict[int, dict[int, set[int]]] = defaultdict(dict)
    for key, value in d.items():
        grouped[len(value)][key] = value

    masks: list[np.ndarray] = []
    data: list[np.ndarray] = []
    t_arrs: list[np.ndarray] = []
    t_hashs: list[np.ndarray] = []
    a_hashs: list[np.ndarray] = []

    for group in list(grouped.values()):
        mask = np.zeros_like(atom_hash, dtype=np.bool_)
        k = [int(i) for i in group.keys()]
        mask[k] = True
        group_values = [(k, *v) for k, v in group.items()]  # rename me

        n_neigh = len(group_values[0])
        perm = itertools.permutations(range(1, n_neigh))

        perm_with_zero = [(0,) + p for p in perm]

        g = np.array(group_values, dtype=np.int64)[:, perm_with_zero]

        masks.append(mask)
        data.append(g)
        t_arrs.append(np.empty_like(g, dtype=np.int64))
        t_hashs.append(np.empty(shape=t_arrs[-1].shape[:-1], dtype=np.int64))
        a_hashs.append(np.empty(shape=t_hashs[-1].shape[:-1], dtype=np.int64))

    n_atom_classes = None
    counter = itertools.repeat(None) if max_iter is None else range(max_iter)
    new_atom_hash = np.empty_like(atom_hash, dtype=np.int64)
    
    for _ in counter:
        for d, m, t_arr, t_hash, a_hash in zip(
            data, masks, t_arrs, t_hashs, a_hashs
        ):
            t_arr[:] = atom_hash[d]
            t_hash = numpy_int_tuple_hash(t_arr, out=t_hash)
            t_hash.sort(axis=-1) # defaults to quicksort
            a_hash = numpy_int_tuple_hash(t_hash, out=a_hash)
            new_atom_hash[m] = a_hash

        new_n_classes = np.unique(new_atom_hash).shape[0]

        if new_n_classes == n_atom_classes:
            break
        elif new_n_classes == n_atoms:
            break
        else:
            n_atom_classes = new_n_classes
            atom_hash, new_atom_hash = new_atom_hash, atom_hash

    return {a: int(h) for a, h in zip(mg.atoms, atom_hash)}




# pyright: standard
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import rdkit.Chem as Chem  # type: ignore

from stereomolgraph.algorithms.bond_orders import connectivity2bond_orders
from stereomolgraph.stereodescriptors import (
    AtropBond,
    Octahedral,
    PlanarBond,
    SquarePlanar,
    Tetrahedral,
    TrigonalBipyramidal,
)

if TYPE_CHECKING:
    from stereomolgraph.graphs.crg import CondensedReactionGraph
    from stereomolgraph.graphs.mg import MolGraph
    from stereomolgraph.graphs.smg import StereoMolGraph


bond_type_dict = {
    0.5: Chem.BondType.HYDROGEN,
    # to be drawn as a dotted line
    # (looks better than other options)
    0: Chem.BondType.UNSPECIFIED,
    1: Chem.BondType.SINGLE,
    2: Chem.BondType.DOUBLE,
    3: Chem.BondType.TRIPLE,
    4: Chem.BondType.QUADRUPLE,
    5: Chem.BondType.QUINTUPLE,
    6: Chem.BondType.HEXTUPLE,
    1.5: Chem.BondType.ONEANDAHALF,
    2.5: Chem.BondType.TWOANDAHALF,
    3.5: Chem.BondType.THREEANDAHALF,
    4.5: Chem.BondType.FOURANDAHALF,
    5.5: Chem.BondType.FIVEANDAHALF,
    "AROMATIC": Chem.BondType.AROMATIC,
    "IONIC": Chem.BondType.IONIC,
    "HYDROGEN": Chem.BondType.HYDROGEN,
    "THREECENTER": Chem.BondType.THREECENTER,
    "DATIVEONE": Chem.BondType.DATIVEONE,
    "DATIVE": Chem.BondType.DATIVE,
    "DATIVEL": Chem.BondType.DATIVEL,
    "DATIVER": Chem.BondType.DATIVER,
    "OTHER": Chem.BondType.OTHER,
    "ZERO": Chem.BondType.ZERO,
}


def set_bond_orders(
    graph: MolGraph,
    mol: Chem.rdchem.RWMol,
    idx_map_num_dict: dict[int, int],
    allow_charged_fragments=False,
    charge=0,
) -> Chem.rdchem.RWMol:
    bond_order_mat, atomic_charges, unpaired_electrons = (
        connectivity2bond_orders(
            atom_types=graph.atom_types,
            connectivity_matrix=graph.connectivity_matrix(),
            allow_charged_fragments=allow_charged_fragments,
            charge=charge,
        )
    )

    index_map_num_dict = {i: map_num for i, map_num in enumerate(graph.atoms)}

    map_num_idx_dict = {
        map_num: idx for idx, map_num in idx_map_num_dict.items()
    }

    for bond in graph.bonds:
        atom1, atom2 = bond
        bond_order = bond_order_mat[index_map_num_dict[atom1]][
            index_map_num_dict[atom2]
        ]

        mol.GetBondBetweenAtoms(
            map_num_idx_dict[atom1], map_num_idx_dict[atom2]
        ).SetBondType(bond_type_dict[bond_order])

    for i, atomic_charge in enumerate(atomic_charges):
        if atomic_charge:
            mol.GetAtomWithIdx(i).SetFormalCharge(int(atomic_charge))
    for i, unpaired_e in enumerate(unpaired_electrons):
        if unpaired_e:
            mol.GetAtomWithIdx(i).SetNumRadicalElectrons(unpaired_e)

    return mol


def mol_graph_to_rdmol(
    graph: MolGraph,
    generate_bond_orders=False,
    allow_charged_fragments=False,
    charge=0,
) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
    mol = Chem.RWMol()

    atom_types_strings = []
    idx_map_num_dict = {}

    for atom in graph.atoms:
        atom_type = graph.get_atom_attribute(atom, "atom_type")
        if atom_type is None:
            raise RuntimeError(atom, graph.atoms, graph.atom_types)
        rd_atom = Chem.Atom(atom_type.symbol)
        rd_atom.SetNoImplicit(True)
        atom_index = mol.AddAtom(rd_atom)
        idx_map_num_dict[atom_index] = atom
        atom_types_strings.append(atom_type.atomic_nr)
        try:
            mol.GetAtomWithIdx(atom_index).SetAtomMapNum(atom)
        except OverflowError as e:
            warnings.warn(f"Atom number number not set: {e}")

    for i in range(graph.n_atoms):
        for j in range(i + 1, graph.n_atoms):
            if graph.has_bond(idx_map_num_dict[i], idx_map_num_dict[j]):
                mol.AddBond(i, j)
                # TODO: check if this is still needed
                # mol.GetBondBetweenAtoms(i, j).SetBondType(
                #    Chem.rdchem.BondType.SINGLE)
    if generate_bond_orders:
        mol = set_bond_orders(
            graph=graph,
            mol=mol,
            idx_map_num_dict=idx_map_num_dict,
            allow_charged_fragments=allow_charged_fragments,
            charge=charge,
        )

    return mol, idx_map_num_dict


def stereo_mol_graph_to_rdmol(
    graph: StereoMolGraph,
    generate_bond_orders=False,
    allow_charged_fragments=False,
    charge=0,
) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
    """
    Creates a RDKit mol object using the connectivity of the mol graph.
    Stereochemistry is added to the mol object.

    :return: RDKit molecule
    :rtype: Chem.rdchem.RWMol
    """
    rd_tetrahedral = {
        1: Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
        -1: Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
        None: Chem.rdchem.ChiralType.CHI_TETRAHEDRAL,
    }
    mol, idx_map_num_dict = mol_graph_to_rdmol(
        graph,
        generate_bond_orders=generate_bond_orders,
        allow_charged_fragments=allow_charged_fragments,
        charge=charge,
    )

    map_num_idx_dict = {v: k for k, v in idx_map_num_dict.items()}

    for atom in graph.atoms:
        a_stereo = graph.get_atom_stereo(atom)
        atom_idx = map_num_idx_dict[atom]
        rd_atom = mol.GetAtomWithIdx(atom_idx)

        if a_stereo is not None and any(
            a not in graph.atoms for a in a_stereo.atoms
        ):
            raise NotImplementedError(
                "Handling of missing atoms not supported yet"
            )
            for mis_a in [
                a for a in a_stereo.atoms[1:] if a not in graph.atoms
            ]:
                if mis_a not in map_num_idx_dict:
                    # add dummy atom
                    idx = mol.AddAtom(Chem.Atom(0))
                    map_num_idx_dict[mis_a] = idx
                    idx_map_num_dict[idx] = mis_a
                    try:
                        mol.GetAtomWithIdx(idx).SetAtomMapNum(mis_a)
                    except OverflowError as e:
                        warnings.warn(f"Atom number number not set: {e}")

                mol.AddBond(
                    map_num_idx_dict[atom],
                    map_num_idx_dict[mis_a],
                    Chem.BondType.SINGLE,
                )

        # The chirality of an Atom in rdkit is determined by two things:
        # 1. its chiralTag
        # 2. the input order of its bonds
        # (see note below for handling of implicit Hs)
        #
        # For tetrahedral coordination, the chiralTag tells you what
        # direction you have to rotate to get from bond 2 to bond 3 while
        # looking down bond 1. This is pretty much identical to the SMILES
        # representation of chirality.

        if a_stereo is not None and isinstance(a_stereo, Tetrahedral):
            mol.GetAtomWithIdx(atom_idx).SetHybridization(
                Chem.HybridizationType.SP3
            )
            assert len(a_stereo.atoms) == 5
            rd_bonds = [
                idx_map_num_dict[a.GetIdx()]
                for a in mol.GetAtomWithIdx(atom_idx).GetNeighbors()
            ]
            perm = [
                p
                for p in a_stereo._perm_atoms()
                if p[1] == rd_bonds[0] and p[2] == rd_bonds[1]
            ][0]

            if (perm[3], perm[4]) == (rd_bonds[2], rd_bonds[3]):
                mol.GetAtomWithIdx(atom_idx).SetChiralTag(
                    rd_tetrahedral[a_stereo.parity]
                )

            elif (perm[3], perm[4]) == (rd_bonds[3], rd_bonds[2]):
                mol.GetAtomWithIdx(atom_idx).SetChiralTag(
                    rd_tetrahedral[a_stereo.parity]
                )
            else:
                raise RuntimeError(
                    "Central atom was not bonded to all Stereo Atoms"
                )

        elif a_stereo is not None and isinstance(a_stereo, SquarePlanar):
            rd_atom.SetChiralTag(Chem.ChiralType.CHI_SQUAREPLANAR)
            neighbors = tuple(
                [
                    idx_map_num_dict[atom.GetIdx()]
                    for atom in rd_atom.GetNeighbors()
                ]
            )

            if neighbors in {p[1:] for p in a_stereo._perm_atoms()}:
                rd_atom.SetUnsignedProp("_chiralPermutation", 1)
            else:
                rd_atom.SetUnsignedProp("_chiralPermutation", 2)

        elif a_stereo is not None and isinstance(
            a_stereo, TrigonalBipyramidal
        ):
            # the order of the neighbors is important for stereochemistry
            for rd_n in rd_atom.GetNeighbors():
                mol.RemoveBond(rd_n.GetIdx(), atom_idx)

            for a in (1, 3, 4, 5, 2):
                a = a_stereo.atoms[a]
                mol.AddBond(
                    atom_idx,
                    map_num_idx_dict[a],
                )
            rd_atom.SetHybridization(Chem.HybridizationType.SP3D)
            rd_atom.SetChiralTag(Chem.ChiralType.CHI_TRIGONALBIPYRAMIDAL)
            if a_stereo.parity == 1:
                rd_atom.SetUnsignedProp("_chiralPermutation", 1)
            elif a_stereo.parity == -1:
                rd_atom.SetUnsignedProp("_chiralPermutation", 2)

        elif a_stereo is not None and isinstance(a_stereo, Octahedral):
            for rd_n in rd_atom.GetNeighbors():
                mol.RemoveBond(rd_n.GetIdx(), atom_idx)

            for a in (1, 5, 6, 3, 4, 2):
                a = a_stereo.atoms[a]
                mol.AddBond(
                    atom_idx,
                    map_num_idx_dict[a],
                )
            rd_atom.SetChiralTag(Chem.ChiralType.CHI_OCTAHEDRAL)
            rd_atom.SetHybridization(Chem.HybridizationType.SP3D2)
            if a_stereo.parity == 1:
                rd_atom.SetUnsignedProp("_chiralPermutation", 1)
            elif a_stereo.parity == -1:
                rd_atom.SetUnsignedProp("_chiralPermutation", 2)

    for b_stereo in (bs for bs in graph.bond_stereo.values() if bs):
        a1, a2 = b_stereo.atoms[2], b_stereo.atoms[3]

        if not all(a in graph.atoms for a in b_stereo.atoms):
            raise NotImplementedError(
                "Handling of missing atoms not supported yet"
            )
            for mis_a in (a for a in b_stereo.atoms if a not in graph.atoms):
                if mis_a not in map_num_idx_dict:
                    # add dummy atom
                    idx = mol.AddAtom(Chem.Atom(0))
                    map_num_idx_dict[mis_a] = idx
                    idx_map_num_dict[idx] = mis_a
                    try:
                        mol.GetAtomWithIdx(idx).SetAtomMapNum(mis_a)
                    except OverflowError as e:
                        warnings.warn(f"Atom number number not set: {e}")

                if mis_a in b_stereo.atoms[:2]:
                    mol.AddBond(
                        map_num_idx_dict[b_stereo.atoms[2]],
                        map_num_idx_dict[mis_a],
                        Chem.BondType.SINGLE,
                    )
                elif mis_a in b_stereo.atoms[4:6]:
                    mol.AddBond(
                        map_num_idx_dict[b_stereo.atoms[3]],
                        map_num_idx_dict[mis_a],
                        Chem.BondType.SINGLE,
                    )
                else:
                    raise RuntimeError("This should not happen")

        rd_a1 = map_num_idx_dict[a1]
        rd_a2 = map_num_idx_dict[a2]
        rd_bond = mol.GetBondBetweenAtoms(rd_a1, rd_a2)
        new_a1 = idx_map_num_dict[rd_bond.GetBeginAtomIdx()]
        new_a2 = idx_map_num_dict[rd_bond.GetEndAtomIdx()]

        assert {a1, a2} == {new_a1, new_a2}

        if isinstance(b_stereo, PlanarBond):
            mol.GetAtomWithIdx(rd_a1).SetHybridization(
                Chem.HybridizationType.SP2
            )
            mol.GetAtomWithIdx(rd_a2).SetHybridization(
                Chem.HybridizationType.SP2
            )

            if b_stereo.parity is None:
                rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREONONE)

            elif (a1, a2) == (new_a1, new_a2):
                rd_bond.SetStereoAtoms(
                    map_num_idx_dict[b_stereo.atoms[0]],
                    map_num_idx_dict[b_stereo.atoms[4]],
                )
                rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOZ)

            elif (a1, a2) == (new_a2, new_a1):
                rd_bond.SetStereoAtoms(
                    map_num_idx_dict[b_stereo.atoms[4]],
                    map_num_idx_dict[b_stereo.atoms[0]],
                )
                rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOZ)
            else:
                raise Exception(f"something wrong with {b_stereo}")

            # if no planar bond neigboring set the bond to double
            if False:
                ...
                # TODO: check if this is still needed
                # all(
                # graph.get_bond_stereo(
                #    (b_stereo.atoms[i], b_stereo.atoms[j])
                # )
                # is None
                # for i, j in ((0, 2), (1, 2), (3, 4), (3, 5))
                # if tuple(sorted((b_stereo.atoms[i], b_stereo.atoms[j])))
                #         in graph.bonds):
        #
        #     rd_bond.SetBondType(Chem.BondType.DOUBLE)

        elif isinstance(b_stereo, AtropBond):
            if (a1, a2) == (new_a1, new_a2):
                rd_bond.SetStereoAtoms(
                    map_num_idx_dict[b_stereo.atoms[0]],
                    map_num_idx_dict[b_stereo.atoms[4]],
                )
                if b_stereo.parity == 1:
                    rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOATROPCW)
                elif b_stereo.parity == -1:
                    rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOATROPCCW)

            elif (a1, a2) == (new_a2, new_a1):
                rd_bond.SetStereoAtoms(
                    map_num_idx_dict[b_stereo.atoms[4]],
                    map_num_idx_dict[b_stereo.atoms[0]],
                )
                if b_stereo.parity == 1:
                    rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOATROPCCW)
                elif b_stereo.parity == -1:
                    rd_bond.SetStereo(Chem.rdchem.BondStereo.STEREOATROPCW)
            else:
                raise Exception(f"something wrong with {b_stereo}")

    if generate_bond_orders:
        mol = set_bond_orders(
            graph=graph,
            mol=mol,
            idx_map_num_dict=idx_map_num_dict,
        )

    return mol, idx_map_num_dict


def set_crg_bond_orders(
    graph: CondensedReactionGraph,
    mol: Chem.rdchem.RWMol,
    idx_map_num_dict: dict[int, int],
    generate_bond_orders=False,
    allow_charged_fragments=False,
    charge=0,
) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
    r, r_idx_map_num_dict = mol_graph_to_rdmol(
        graph.reactant(),
        generate_bond_orders=True,
        allow_charged_fragments=allow_charged_fragments,
        charge=charge,
    )
    p, p_idx_map_num_dict = mol_graph_to_rdmol(
        graph.product(),
        generate_bond_orders=True,
        allow_charged_fragments=allow_charged_fragments,
        charge=charge,
    )
    assert r_idx_map_num_dict == p_idx_map_num_dict == idx_map_num_dict

    for bond in mol.GetBonds():
        a1, a2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        if (r_bond := r.GetBondBetweenAtoms(a1, a2)) and (
            p_bond := p.GetBondBetweenAtoms(a1, a2)
        ):
            r_bond_order = r_bond.GetBondTypeAsDouble()
            p_bond_order = p_bond.GetBondTypeAsDouble()
            average = (r_bond_order + p_bond_order) / 2
            bond_order = round(average * 2) / 2
            mol.GetBondBetweenAtoms(a1, a2).SetBondType(
                bond_type_dict[bond_order]
            )

    return mol



from __future__ import annotations

import itertools
from collections import deque
from typing import TYPE_CHECKING

from stereomolgraph import StereoCondensedReactionGraph, StereoMolGraph
from stereomolgraph.graphs.scrg import Change

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable
    from typing import Optional

    from stereomolgraph.graphs.smg import AtomId, Bond, StereoMolGraph


def generate_stereoisomers(
    graph: StereoMolGraph,
    enantiomers: bool = True,
    atoms: Optional[Iterable[AtomId]] = None,
    bonds: Optional[Iterable[Bond]] = None,
) -> Collection[StereoMolGraph]:
    """Generates all stereoisomers of a StereoMolGraph by generation of
    all combinations of parities. Only includes stereocenter which have a
    parity of None. If a parity is set, it is not changed.

    If include_enantiomers is True, both enantiomers of a stereoisomer are
    included, if it is False, only one enantiomer is included.

    :param enantiomers: If True, both enantiomers are included,
    :param: sets if both enantiomers should be included, default: Ture
    :return: All possible stereoisomers
    """
    if atoms is None:
        atom_stereos = (
            stereo.get_isomers()
            for a in graph.atoms
            if ((stereo := graph.get_atom_stereo(a)) and stereo.parity is None)
        )
    else:
        atom_stereos = (
            stereo.get_isomers()
            for a in atoms
            if (stereo := graph.get_atom_stereo(a)) is not None
        )

    if bonds is None:
        bond_stereos = (
            stereo.get_isomers()
            for b in graph.bonds
            if ((stereo := graph.get_bond_stereo(b)) and stereo.parity is None)
        )
    else:
        bond_stereos = (
            stereo.get_isomers()
            for b in bonds
            if (stereo := graph.get_bond_stereo(b)) is not None
        )

    isomers: set[StereoMolGraph] = set()
    enantiomers_set: set[StereoMolGraph] = set()

    for a_stereos, b_stereos in itertools.product(
        itertools.product(*atom_stereos), itertools.product(*bond_stereos)
    ):
        stereoisomer = graph.copy()
        for a_stereo in a_stereos:
            stereoisomer.set_atom_stereo(a_stereo)
        for b_stereo in b_stereos:
            stereoisomer.set_bond_stereo(b_stereo)

        if stereoisomer not in enantiomers_set:
            isomers.add(stereoisomer)

            if not enantiomers:
                enantiomers_set.add(stereoisomer.enantiomer())

    return isomers


def generate_fleeting_stereoisomers(
    graph: StereoCondensedReactionGraph,
    enantiomers: bool = True,
    atoms: Optional[Iterable[AtomId]] = None,
    bonds: Optional[Iterable[Bond]] = None,
) -> Collection[StereoCondensedReactionGraph]:
    # TODO: extend to more than fleeting stereochemistry
    # add checks if fleeting stereochemistry is valid relative to formed
    # and broken

    if atoms is None:
        atom_stereos = [
            stereo.get_isomers()
            for a in graph.atoms
            if (
                (stereo_change_dict := graph.get_atom_stereo_change(a))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        ]
    else:
        atom_stereos = [
            stereo.get_isomers()
            for a in atoms
            if (
                (stereo_change_dict := graph.get_atom_stereo_change(a))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        ]

    if bonds is None:
        bond_stereos = [
            stereo.get_isomers()
            for b in graph.bonds
            if (
                (stereo_change_dict := graph.get_bond_stereo_change(b))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        ]
    else:
        bond_stereos = [
            stereo.get_isomers()
            for b in bonds
            if (
                (stereo_change_dict := graph.get_bond_stereo_change(b))
                and (stereo := stereo_change_dict[Change.FLEETING])
            )
            and stereo.parity is None
        ]

    isomers: list[StereoCondensedReactionGraph] = []
    enantiomers_set: set[StereoCondensedReactionGraph] = set()

    for a_stereos, b_stereos in itertools.product(
        itertools.product(*atom_stereos), itertools.product(*bond_stereos)
    ):
        stereoisomer = graph.copy()
        for a_stereo in a_stereos:
            stereoisomer.set_atom_stereo_change(fleeting=a_stereo)
        for b_stereo in b_stereos:
            stereoisomer.set_bond_stereo_change(fleeting=b_stereo)

        if stereoisomer not in enantiomers_set:
            isomers.append(stereoisomer)

            if not enantiomers:
                enantiomers_set.add(stereoisomer.enantiomer())

    return isomers


def topological_symmetry_number(graph: StereoMolGraph) -> int:
    """
    Calculated from the number of graph isomorphisms which conserve the
    stereo information.
    symmetry_number = internal_symmetry_number * rotational_symmetry_number
    TODO: add paper reference
    """

    if any(stereo.parity is None for stereo in graph.stereo.values()):
        raise NotImplementedError(
            "all stereocenters have to be defined"
            " to calculate the symmetry number"
        )

    mappings = graph.get_isomorphic_mappings(graph)
    return deque(enumerate(mappings, 1), maxlen=1)[0][0]



from __future__ import annotations

from collections import Counter, defaultdict
from copy import deepcopy
from types import MappingProxyType
from typing import TYPE_CHECKING

import numpy as np

from stereomolgraph import PERIODIC_TABLE, Element
from stereomolgraph.coords import BondsFromDistance
from stereomolgraph.algorithms.isomorphism import vf2pp_all_isomorphisms
from stereomolgraph.algorithms.color_refine import color_refine_mg
from stereomolgraph.graph2rdmol import mol_graph_to_rdmol
from stereomolgraph.rdmol2graph import mol_graph_from_rdmol

if TYPE_CHECKING:

    from collections.abc import Iterable, Iterator, Mapping, Sequence
    from typing import Any, Optional, TypeAlias, Self, TypeVar
    
    from rdkit import Chem # type: ignore

    from stereomolgraph.coords import Geometry
    N = TypeVar("N", bound=int)

AtomId: TypeAlias = int

Bond: TypeAlias = frozenset[AtomId]


class MolGraph:
    """
    Graph representing a molecular entity. Nodes represent atoms and edges
    represent bonds. All nodes have an `atom_type` attribute of type `Element`.
    The node ids should be integers. The graph is considered equal to another
    graph, iff. they are isomorphic and of the same type.
    """
    __slots__ = ("_atom_attrs", "_neighbors", "_bond_attrs")

    _atom_attrs: dict[AtomId, dict[str, Any]]
    _neighbors: dict[AtomId, set[AtomId]]
    _bond_attrs: dict[Bond, dict[str, Any]]

    def __init__(self, mol_graph: Optional[MolGraph] = None):
        if mol_graph is not None:
            self._atom_attrs = deepcopy(mol_graph._atom_attrs)
            self._neighbors = deepcopy(mol_graph._neighbors)
            self._bond_attrs = deepcopy(mol_graph._bond_attrs)
        else:
            self._atom_attrs = defaultdict(dict)
            self._neighbors = defaultdict(set)
            self._bond_attrs = defaultdict(dict)

    @property
    def atoms(
        self,
    ) -> Iterable[AtomId]:
        """
        :return: Returns all atoms of the molecule
        """
        return self._atom_attrs.keys()

    @property
    def atom_types(
        self,
    ) -> tuple[Element, ...]:
        """
        :return: Returns all atom types in the MolGraph
        """
        return tuple([v["atom_type"] for v in self._atom_attrs.values()])

    @property
    def atoms_with_attributes(self) -> Mapping[AtomId, dict[str, Any]]:
        """
        :return: Returns all atoms in the MolGraph with their attributes
        """
        return MappingProxyType(self._atom_attrs)

    @property
    def bonds(
        self,
    ) -> Iterable[Bond]:
        """
        :return: Returns all bonds in the MolGraph
        """
        return self._bond_attrs.keys()

    @property
    def bonds_with_attributes(
        self,
    ) -> Mapping[Bond, dict[str, Any]]:
        """
        :return: Returns all bonds in the MolGraph with their attributes
        """
        return MappingProxyType(self._bond_attrs)

    @property
    def neighbors(
        self,
    ) -> Mapping[AtomId, set[AtomId]]:
        """
        :return: Returns all neighbors of the atoms in the MolGraph
        """
        return MappingProxyType(self._neighbors)

    @property
    def n_atoms(
        self,
    ) -> int:
        """
        :return: Returns number of atoms in the MolGraph
        """
        return len(self._atom_attrs)

    def __len__(self) -> int:
        return len(self._atom_attrs)

    def has_atom(self, atom: int) -> bool:
        """Returns True if the molecules contains an atom with this id.

        :param atom: Atom
        :return: value
        """
        return atom in self._atom_attrs

    def add_atom(
        self, atom: AtomId, atom_type: int | str | Element,
        **attr: Any
    ):
        """Adds atom to the MolGraph

        :param atom: Atom ID
        :param atom_type: Atom Type
        """
        atom_type = PERIODIC_TABLE[atom_type]

        self._atom_attrs[atom] = {"atom_type": atom_type, **attr}

    def remove_atom(self, atom: AtomId):
        """Removes atom from graph.
        
        :param atom: Atom ID
        :raises: KeyError if atom is not in graph.
        """
        del self._atom_attrs[atom]
        if nbr := self._neighbors.pop(atom, None):
            for n in nbr:
                self.remove_bond(atom, n)

    def get_atom_attribute(
        self, atom: AtomId, attr: str
    ) -> Optional[Any]:
        """
        Returns the value of the attribute of the atom or None if the atom does
        not have this attribute.
        Raises KeyError if atom is not in graph.

        :param atom: Atom
        :param attr: Attribute
        :raises KeyError: Atom not in graph
        :return: Returns the value of the attribute of the atom
        """
        return self._atom_attrs[atom].get(attr, None)

    def get_atom_type(self, atom: AtomId) -> Element:
        """
        Returns the atom type of the atom.
        Raises KeyError if atom is not in graph.

        :param atom: Atom
        :raises KeyError: Atom not in graph
        :return: Returns the atom type of the atom
        """
        return self._atom_attrs[atom]["atom_type"]

    def set_atom_attribute(self, atom: AtomId, attr: str, value: Any):
        """
        sets the Value of the Attribute on Atom.
        Raises KeyError if atom is not in graph.

        :param atom: Atom
        :param attr: Attribute
        :param value: Value
        :raises KeyError: Atom not in graph
        :raises ValueError: The attribute "atom_type" can only have values of
                            type Element
        """
        if attr == "atom_type":
            try:
                value = PERIODIC_TABLE[value]
            except KeyError:
                raise ValueError(
                    f"'{value}' can not be used as atom_type for "
                    f"{self.__class__.__name__}"
                )
        self._atom_attrs[atom][attr] = value

    def delete_atom_attribute(self, atom: AtomId, attr: str):
        """
        Deletes the Attribute of the Atom
        Raises KeyError if attribute is not present.
        Raises KeyError if atom is not in graph.

        :param atom: Atom ID
        :param attr: Attribute
        :raises ValueError: The attribute "atom_type" can not be deleted
        """
        if attr == "atom_type":
            raise ValueError("atom_type can not be deleted")
        else:
            self._atom_attrs[atom].pop(attr)

    def get_atom_attributes(
        self, atom: AtomId, attributes: Optional[Iterable[str]] = None
    ) -> Mapping[str, Any]:
        """
        Returns the attributes of the atom. If no attributes are given, all
        attributes are returned.
        Raises KeyError if atom is not in graph.

        :param atom: Atom
        :param attributes: Specific attributes to return
        :return: Returns all or just the chosen attributes of the atom
        """
        if attributes is None:
            return MappingProxyType(self._atom_attrs[atom])
        else:
            return {attr: self._atom_attrs[atom][attr] for attr in attributes}

    def has_bond(self, atom1: AtomId, atom2: AtomId) -> bool:
        """Returns True if bond is in MolGraph.

        :param atom1: Atom1
        :param atom2: Atom2
        :return: If the bond is in MolGraph
        """
        return Bond({atom1, atom2}) in self._bond_attrs

    def add_bond(self, atom1: AtomId, atom2: AtomId,
                 **attr: Any):
        """Adds bond between Atom1 and Atom2.

        :param atom1: Atom1
        :param atom2: Atom2
        """
        if atom1 not in self.atoms or atom2 not in self.atoms:
            raise ValueError("Atoms not in Graph")
        bond = Bond({atom1, atom2})
        self._neighbors[atom1].add(atom2)
        self._neighbors[atom2].add(atom1)
        self._bond_attrs[bond] = attr

    def remove_bond(self, atom1: AtomId, atom2: AtomId):
        """
        Removes bond between Atom1 and Atom2.

        :param atom1: Atom1
        :param atom2: Atom2
        """
        bond = Bond((atom1, atom2))
        del self._bond_attrs[bond]
        self._neighbors[atom1].discard(atom2)
        self._neighbors[atom2].discard(atom1)

    def get_bond_attribute(
        self, atom1: AtomId, atom2: AtomId, attr: str,
    ) -> Any:
        """
        Returns the value of the attribute of the bond between Atom1 and Atom2.
        Raises KeyError if bond is not in graph.

        :param atom1: Atom1
        :param atom2: Atom2
        :param attr: Attribute
        :return: Returns the value of the attribute of the bond
                 between Atom1 and Atom2
        """
        bond = Bond((atom1, atom2))
        if bond in self._bond_attrs:
            return self._bond_attrs[bond].get(attr, None)
        else:
            raise ValueError(f"No Bond between {atom1} and {atom2}")

    def set_bond_attribute(
        self, atom1: AtomId, atom2: AtomId, attr: str, value: Any
    ):
        """
        sets the Attribute of the bond between Atom1 and Atom2.
        The Attribute "bond_order" can only have numerical values.
        Raises KeyError if bond is not in graph.

        :param atom1: Atom1
        :param atom2: Atom2
        :param attr: Attribute
        :param value: Value
        """
        bond = Bond((atom1, atom2))
        if bond in self._bond_attrs:
            self._bond_attrs[bond][attr] = value
        else:
            raise ValueError(f"No Bond between {atom1} and {atom2}")

    def delete_bond_attribute(self, atom1: AtomId, atom2: AtomId,
                              attr: str):
        """
        Deletes the Attribute of the bond between Atom1 and Atom2
        
        :param atom1:
        :param atom2: Atom1
        :param attr: Attribute
        """
        self._bond_attrs[Bond((atom1, atom2))].pop(attr)

    def get_bond_attributes(
        self,
        atom1: AtomId,
        atom2: AtomId,
        attributes: Optional[Iterable[str]] = None,
    ) -> Mapping[str, Any]:
        """
        :param atom1: Atom1
        :param atom2: Atom2
        :param attributes: Specific attributes to return
        :return: Returns chosen attributes of the bond between Atom1 and Atom2
        """
        bond = Bond((atom1, atom2))
        if attributes is None:
            return MappingProxyType(self._bond_attrs[bond])
        else:
            return {attr: val for attr, val in self._bond_attrs[bond].items()}

    def bonded_to(self, atom: int) -> frozenset[int]:
        """
        Returns the atoms connected to the atom.

        :param atom: Id of the atom.
        :return: tuple of atoms connected to the atom.
        """
        return frozenset(self._neighbors[atom])

    def connectivity_matrix(self) -> np.ndarray[tuple[N, N], np.dtype[np.int8]]:
        """
        Returns a connectivity matrix of the graph as a list of lists.
        Order is the same as in self.atoms()
        1 if nodes are connected, 0 if not.

        :return: Connectivity matrix as list of lists
        """
        matrix = [[0] * self.n_atoms for _ in self.atoms]
        atomid_index_dict = {id:index for index, id in enumerate(self.atoms)}

        for a1, a2 in self.bonds:
            matrix[atomid_index_dict[a1]][atomid_index_dict[a2]] = 1
            matrix[atomid_index_dict[a2]][atomid_index_dict[a1]] = 1
        return np.array(matrix, dtype=np.int8)

    def _to_rdmol(
        self,
        generate_bond_orders: bool=False,
        allow_charged_fragments: bool = False,
        charge: int = 0
    ) -> tuple[Chem.rdchem.RWMol, dict[int, int]]:
        return mol_graph_to_rdmol(self,
                                   generate_bond_orders=generate_bond_orders,
                                   allow_charged_fragments = allow_charged_fragments,
                                   charge=charge)

    def to_rdmol(
        self,
        generate_bond_orders:bool=False,
        allow_charged_fragments:bool = False,
        charge:int=0
    ) -> Chem.rdchem.Mol:

        mol, _ = self._to_rdmol(
            generate_bond_orders=generate_bond_orders,
            allow_charged_fragments = allow_charged_fragments,
            charge=charge
        )
        for atom in mol.GetAtoms(): # type: ignore
            atom.SetAtomMapNum(0, strict=True) # type: ignore
        return mol

    @classmethod
    def from_rdmol(cls, rdmol: Chem.Mol, use_atom_map_number:bool=False) -> Self:
        """
        Creates a StereoMolGraph from an RDKit Mol object.
        Implicit Hydrogens are added to the graph.
        Stereo information is conserved. Double bonds, aromatic bonds and
        conjugated bonds are interpreted as planar. Atoms with 5 bonding
        partners are assumed to be TrigonalBipyramidal and allow interchange
        of the substituents (berry pseudorotation). Atoms with 6 bonding
        partners are assumed to be octahedral and do not allow interchange of
        the substituents.

        :param rdmol: RDKit Mol object
        :param use_atom_map_number: If the atom map number should be used
                                    instead of the atom index
        :return: StereoMolGraph
        """
        return mol_graph_from_rdmol(cls, rdmol, use_atom_map_number=use_atom_map_number) # type: ignore

    def relabel_atoms(
        self, mapping: dict[int, int], copy: bool = True
    ) -> Self:
        """Changes the atom labels according to mapping.

        :param mapping: dict used for map old atom labels to new atom labels
        :param copy: defines if the relabeling is done inplace or a new object
                     should be created
        :return: this object (self) or a new instance of self.__class__
        """
        atom_attrs = {
            mapping.get(atom, atom): attrs
            for atom, attrs in self._atom_attrs.items()
        }
        neighbors = {
            mapping.get(atom, atom): {mapping.get(n, n) for n in neighbors}
            for atom, neighbors in self._neighbors.items()
        }

        bond_attrs = {
            Bond({mapping.get(atom, atom) for atom in bond}): attrs
            for bond, attrs in self._bond_attrs.items()
        }
        if copy is True:
            new_graph = self.__class__()
        elif copy is False:
            new_graph = self

        new_graph._atom_attrs = atom_attrs
        new_graph._neighbors = neighbors
        new_graph._bond_attrs = bond_attrs
        return new_graph

    def node_connected_component(self, atom: int) -> set[AtomId]:
        """
        :param atom: atom id
        :return: Returns the connected component that includes atom_id
        """
        visited: set[AtomId] = set()
        stack = [atom]
        while stack:
            node = stack.pop()

            if node not in visited:
                visited.add(node)
            for neighbor in self.bonded_to(node):
                if neighbor not in visited:
                    stack.append(neighbor)
        return visited

    def connected_components(self) -> list[set[int]]:
        """
        :return: Returns the connected components of the graph
        """
        visited: set[AtomId] = set()
        components: list[set[int]] = []

        for atom in self.atoms:
            if atom not in visited:
                component = self.node_connected_component(atom)
                components.append(component)
                visited.update(component)

        return components

    def subgraph(self, atoms: Iterable[AtomId]) -> Self:
        """
        Returns a subgraph copy only containing the given atoms

        :param atoms: Iterable of atom ids to be
        :return: Subgraph
        """
        new_atoms = set(atoms)
        atom_attrs = {atom: self._atom_attrs[atom] for atom in atoms}
        bond_attrs = {
            bond: attrs
            for bond, attrs in self._bond_attrs.items()
            if new_atoms.issuperset(bond)
        }
        neighbors = {
            atom: {n for n in self._neighbors[atom] if n in new_atoms}
            for atom in new_atoms
        }
        new_graph = self.__class__()
        new_graph._atom_attrs = atom_attrs
        new_graph._neighbors = neighbors
        new_graph._bond_attrs = bond_attrs
        return new_graph

    def copy(self) ->Self:
        """
        :return: returns a copy of self
        """
        return deepcopy(self)

    def bonds_from_bond_order_matrix(
        self,
        matrix: np.ndarray,
        threshold: float = 0.5,
        include_bond_order: bool = False,
    ):
        """
        Adds bonds the the graph based on bond orders from a matrix

        :param matrix: Bond order Matrix
        :param threshold: Threshold for bonds to be included as edges,
                          defaults to 0.5
        :param include_bond_order: If bond orders should be included as edge
                                   attributes, defaults to False
        """

        if not np.shape(matrix) == (len(self), len(self)):
            raise ValueError(
                "Matrix has the wrong shape. shape of matrix is "
                f"{np.shape(matrix)}, but {len(self), len(self)} "
                "expected"
            )

        bonds = (matrix > threshold).nonzero()

        for i, j in zip(*bonds):
            if include_bond_order:
                self.add_bond(int(i), int(j), bond_order=matrix[i, j])
            else:
                self.add_bond(int(i), int(j))

    @classmethod
    def compose(cls, mol_graphs: Iterable[MolGraph]) -> Self:
        """
        Combines all graphs in the iterable into one. Duplicate nodes or edges
        are overwritten, such that the resulting graph only contains one node
        or edge with that name. Duplicate attributes of duplicate nodes or
        edges are also overwritten in order of iteration.

        :param molgraphs: Iterable of MolGraph that will be composed into a
            single MolGraph
        """
        new_graph = cls()
        for mol_graph in mol_graphs:
            new_graph._atom_attrs.update(mol_graph._atom_attrs)
            new_graph._bond_attrs.update(mol_graph._bond_attrs)

            for atom, neighbors in mol_graph._neighbors.items():
                new_graph._neighbors[atom].update(neighbors)

        return new_graph

    @classmethod
    def from_atom_types_and_bond_order_matrix(
        cls,
        atom_types: Sequence[int | Element | str],
        matrix: np.ndarray,
        threshold:float=0.5,
        include_bond_order:bool=False,
    ):
        """

        :param atom_types: list of atom types as integers or symbols,
                           must correspond to the matrix
        :param matrix: np.matrix of bond orders or connectivities ([0..1])
        :param threshold: Threshold for bonds to be included as edges,
                          defaults to 0.5
        :param include_bond_order: If bond orders should be included as edge
                                   attributes, defaults to False
        :return: Returns MolGraph
        """
        if not len(atom_types) == np.shape(matrix)[0] == np.shape(matrix)[1]:
            raise ValueError(
                "atom_types and matrix have to have the same length"
            )
        new_mol_graph = cls()

        for i, atom_type in enumerate(atom_types):
            new_mol_graph.add_atom(i, atom_type=atom_type)

        x_ids, y_ids = np.triu_indices(matrix.shape[0], k=1)

        # Iterate over the upper triangular matrix excluding the diagonal
        for x_id, y_id in zip(x_ids, y_ids):
            if (value := matrix[x_id, y_id]) >= threshold:
                if include_bond_order is False:
                    new_mol_graph.add_bond(int(x_id), int(y_id))
                elif include_bond_order is True:
                    new_mol_graph.add_bond(
                        int(x_id), int(y_id), bond_order=value
                    )

        return new_mol_graph

    @classmethod
    def from_geometry_and_bond_order_matrix(
        cls,
        geo: Geometry,
        matrix: np.ndarray,
        threshold: float = 0.5,
        include_bond_order: bool = False,
    ) -> Self:
        """
        Creates a graph of a molecule from a Geometry and a bond order matrix.

        :param geo: Geometry
        :param matrix: Bond order matrix
        :param threshold: Threshold for bonds to be included as edges,
            defaults to 0.5
        :param include_bond_order: If bond orders should be included as edge
            attributes, defaults to False
        :return: Graph of Molecule
        """
        new_mol_graph = cls.from_atom_types_and_bond_order_matrix(
            geo.atom_types,
            matrix,
            threshold=threshold,
            include_bond_order=include_bond_order,
        )
        return new_mol_graph

    @classmethod
    def from_geometry(
        cls,
        geo: Geometry,
        switching_function: BondsFromDistance = BondsFromDistance(),
    ) -> Self:
        """
        Creates a graph of a molecule from a Geometry and a switching Function.
        Uses the Default switching function if none are given.

        :param geo: Geometry
        :param switching_function: Function to determine if two atoms are
            connected
        :return: graph of molecule
        """

        connectivity_matrix = switching_function.array(
            geo.coords, geo.atom_types
        )
        return cls.from_geometry_and_bond_order_matrix(
            geo,
            connectivity_matrix,
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.is_isomorphic(other)
        return NotImplemented

    def __hash__(self) -> int:
        color_dict: dict[int, int] = color_refine_mg(self, )
        return hash(frozenset(Counter(color_dict.values()).items()))

    def get_subgraph_isomorphic_mappings(
        self, other: Self,
    ) -> Iterator[dict[int, int]]:
        """Subgraph isomorphic mappings from "other" onto "self".

        Generates all node-iduced subgraph isomorphic mappings.
        All atoms of "other" have to be present in "self".
        The bonds of "other" have to be the subset of the bonds of "self"
        relating to the nodes of "other".

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        """
        return vf2pp_all_isomorphisms(
            self,
            other,
            color_refine=False,
            stereo=False,
            stereo_change=False,
            subgraph=True,
        )

    def get_isomorphic_mappings(self, other: Self) -> Iterator[dict[int, int]]:
        """Isomorphic mappings between "self" and "other".

        Generates all isomorphic mappings between "other" and "self".
        All atoms and bonds have to be present in both graphs.

        :param other: Other Graph to compare with
        :return: Mappings from the atoms of self onto the atoms of other
        :raises TypeError: Not defined for objects different types
        """

        return vf2pp_all_isomorphisms(
            self,
            other,
            color_refine=True, # TODO: implement color refinement
            stereo=False,
            stereo_change=False,
            subgraph=False,
        )

    def is_isomorphic(self, other: Self) -> bool:
        """
        Checks if the graph is isomorphic to another graph.

        :param other: other graph
        :return: True if isomorphic
        """
        return any(self.get_isomorphic_mappings(other))

    def __repr__(self) -> str:
        a_list = [(a, a_type.symbol) for a, a_type
                  in zip(self.atoms, self.atom_types)]
        b_list = [tuple(sorted(bond)) for bond in self.bonds]
        return f"{self.__class__.__name__}\nAtoms: {a_list}\nBonds: {b_list}\n"

    def _ipython_display_(self) -> None:
        print(self.__repr__())
